##########################################################################################
# pdstemplate/__init__.py
##########################################################################################
"""PDS Ring-Moon Systems Node, SETI Institute

``pdstemplate`` is a Python module that defines the :class:`PdsTemplate` class, which is
used to generate PDS labels based on template files. Both PDS3 and PDS4 (xml) labels are
supported. Although specifically designed to facilitate data deliveries by PDS data
providers, the template system is generic and can be used to generate files from templates
for other purposes.

###############
Getting Started
###############

The general procedure is as follows:

1. Create a template object by calling the :meth:`PdsTemplate` constructor to read a
template file::

    from pdstemplate import PdsTemplate
    template = PdsTemplate(template_file_path)

2. Create a dictionary that contains the parameter values to use inside the label.

3. Construct the label using method :meth:`~PdsTemplate.write` as follows::

    template.write(dictionary, label_file)

This will create a new label of the given name, using the values in the given dictionary.
Once the template has been constructed, steps 2 and 3 can be repeated any number of times.

Alternatively, you can obtain the content of a label without writing it to a file using
method :meth:`~PdsTemplate.generate`.

``pdstemplate`` employs the RMS Node's `rms-filecache
<https://pypi.org/project/rms-filecache>`_ module and its `FCPath
<https://rms-filecache.readthedocs.io/en/latest/module.html#filecache.file_cache_path.FCPath>`_
class to support the handling of files at a website or in the cloud. You can refer to a
remote file by URL and the :class:`PdsTemplate` will treat it as if it were a local file.
See `filecache's documentation
<https://rms-filecache.readthedocs.io/en/latest/index.html>`_ for further details.

###############
Template Syntax
###############

A template file will look generally like a label file, except for certain embedded
expressions that will be replaced when the template's :meth:`~PdsTemplate.write` or
:meth:`~PdsTemplate.generate` method is called.

*************
Substitutions
*************

In general, everything between dollar signs "$" in the template is interpreted as a
Python expression to be evaluated. The result of this expression then replaces it
inside the label. For example, if ``dictionary['INSTRUMENT_ID'] == 'ISSWA'``, then::

    <instrument_id>$INSTRUMENT_ID$</instrument_id>

in the template will become::

    <instrument_id>ISSWA</instrument_id>

in the label. The expression between "$" in the template can include indexes, function
calls, or just about any other Python expression. As another example, using the same
dictionary above::

    <camera_fov>$"Narrow" if INSTRUMENT_ID == "ISSNA" else "Wide"$</camera_fov>

in the template will become this in the label::

    <camera_fov>Wide</camera_fov>

An expression in the template of the form ``$name=expression$``, where the `name` is a
valid Python variable name, will also also have the side-effect of defining this
variable so that it can be re-used later in the template. For example, if this appears
as an expression::

    $cruise_or_saturn=('cruise' if START_TIME < 2004 else 'saturn')$

then later in the template, one can write::

    <lid_reference>
    urn:nasa:pds:cassini_iss_$cruise_or_saturn$:data_raw:cum-index
    </lid_reference>

To embed a literal "$" inside a label, enter "$$" into the template.

*******
Headers
*******

Headers provide even more sophisticaed control over the content of a label. A header
appears alone on a line of the template and begins with "$" as the first non-blank
character. It determines whether or how subsequent text of the template will appear in the
file, from here up to the next header line.

===============
FOR and END_FOR
===============

You can include one or more repetitions of the same text using ``FOR`` and ``END_FOR``.
The format is::

    $FOR(expression)
        <template text>
    $END_FOR

where `expression` evaluates to a Python iterable. Within the `template text`, these new
variable names are assigned:

- `VALUE` = the current value of the iterator;
- `INDEX` = the index of this iteration, starting from zero;
- `LENGTH` = the total number of iterations.

For example, if::

    dictionary["targets"] = ["Jupiter", "Io", "Europa"]
    dictionary["naif_ids"] = [599, 501, 502]

then::

    $FOR(targets)
        <target_name>$VALUE (naif_ids[INDEX])$</target_name>
    $END_FOR

in the template will become this in the label::

    <target_name>Jupiter (599)</target_name>
    <target_name>Io (501)</target_name>
    <target_name>Europa (502)</target_name>

Instead of using the names `VALUE`, `INDEX`, and `LENGTH`, you can customize the variable
names by listing up to three comma-separated names and an equal sign "=" before the
iterable expression. For example, this will produce the same results as the example
above::

    $FOR(name, k=targets)
        <target_name>$name (naif_ids[k])$</target_name>
    $END_FOR

=============================
IF, ELSE_IF, ELSE, and END_IF
=============================

You can use ``IF``, ``ELSE_IF``, ``ELSE``, and ``END_IF`` to select among alternative
blocks of text in the template:

- ``IF(expression)``: Evaluate `expression` and include the next lines of the template if
  it is logically True (e.g., boolean True, a nonzero number, a non-empty list or string,
  etc.).

- ``ELSE_IF(expression)``: Include the next lines of the template if `expression` is
  logically True and every previous expression was logically false.

- ``ELSE``: Include the next lines of the template only if all prior expressions were
  logically False.

- ``END_IF``:  This marks the end of the set of if/else alternatives.

As with other substitutions, you can define a new variable of a specified name by using
`name=expression` inside the parentheses of ``IF()`` and ``ELSE_IF()``.

Note that headers can be nested arbitrarily inside the template.

.. _ONCE:

====
ONCE
====

``ONCE`` is a header that simply includes the content that follows it one time. However,
it is useful for its side-effect, which is that ``ONCE(expression)`` allows the embedded
`expression` to be evaluated without writing new text into the label. You can use this
capability to define variables internally without affecting the content of the label
produced. For example::

    $ONCE(date = big_dictionary["key"]["date"])

will assign the value of the variable named `date` for subsequent use within the template.

=======
INCLUDE
=======

This header will read the content of another file and insert its content into the template
here::

    $INCLUDE(filename)

Using the environment variable ``PDSTEMPLATE_INCLUDES``, you can define one or more
directories that will be searched for a file to be included. If multiple directories are
to be searched, they should be separated by colons. You can also specify one or more
directories to search in the :meth:`PdsTemplate` constructor using the `includes` input
parameter.

Include files are handled somewhat differently from other headers. When ``INCLUDE``
references a file as a literal string rather than as an expression to evaluate, it is
processed at the time that the :class:`PdsTemplate` is constructed. However, if the
filename is given as an expression, it is not evaluated until :meth:`~PdsTemplate.write``
or :meth:`~PdsTemplate.generate`` is called for each label.

=================
NOTE and END_NOTE
=================

You can use ``NOTE`` and ``END_NOTE`` to embed any arbitrary comment block into the
template. Any text between these headers does not appear in the label::

    $NOTE
    Here is an extended comment about the templae
    $END_NOTE

You can also use ``$NOTE:`` for an in-line comment. This text, and any blanks before it,
are not included in the label::

    <filter>$FILTER$</filter>   $NOTE: This is where we identify the filter

*********************
Pre-defined Functions
*********************

The following pre-defined functions can be used inside any expression in the template.

- :meth:`~.PdsTemplate.BASENAME`:
  The basename of a filepath, with leading directory path removed.

- :meth:`~.PdsTemplate.BOOL`
  Return one of two strings based on a boolean input.

- :meth:`~.PdsTemplate.COUNTER`
  The current value of a counter.

- :meth:`~.PdsTemplate.CURRENT_TIME`
  The current date or time in the local time zone as a string of the form "yyyy-mm-dd" or
  "yyyy-mm-ddThh:mm:sss".

- :meth:`~.PdsTemplate.CURRENT_ZULU`
  The current UTC date or time as a string of the form "yyyy-mm-dd" or
  "yyyy-mm-ddThh:mm:sssZ".

- :meth:`~.PdsTemplate.DATETIME`
  Convert a time to an ISO time string with the date expressed as "yyyy-mm-dd".

- :meth:`~.PdsTemplate.DATETIME_DOY`
  Convert a time to an ISO time string with the date expressed as "yyyy-ddd".

- :meth:`~.PdsTemplate.DAYSECS`
  Convert a time to the number of elapsed seconds since the most recent midnight.

- :meth:`~.PdsTemplate.FILE_BYTES`
  The size in bytes of a specified file.

- :meth:`~.PdsTemplate.FILE_MD5`
  The MD5 checksum of a specified file.

- :meth:`~.PdsTemplate.FILE_RECORDS`
  The number of records in a specified file.

- :meth:`~.PdsTemplate.FILE_TIME`
  The modification time in the local time zone of the specified file.

- :meth:`~.PdsTemplate.FILE_ZULU`
  The UTC modification time of a specified by file.

- :meth:`~.PdsTemplate.GETENV`
  The value of any environment variable.

- :meth:`~.PdsTemplate.LABEL_PATH`
  The file path of the label being written.

- :meth:`~.PdsTemplate.LOG`
  Write a message to the current log.

- :meth:`~.PdsTemplate.NOESCAPE`
  If the template is XML, evaluated expressions are "escaped" to ensure that they are
  suitable for embedding in a PDS4 label. For example, ">" inside a string will be
  replaced by "&gt;". This function prevents text from being escaped in the label,
  allowing it to contain literal XML.

- :meth:`~.PdsTemplate.QUOTE_IF`
  Quote the given text if it requires quotes within a PDS3 label.

- :meth:`~.PdsTemplate.RAISE`
  Raise an exception with a given class and text message.

- :meth:`~.PdsTemplate.REPLACE_NA`
  Return either a given string or an indication that it is "not applicable".

- :meth:`~.PdsTemplate.REPLACE_UNK`
  Return either a given string or an indication that it is "unknown".

- :meth:`~.PdsTemplate.TEMPLATE_PATH`
  The directory path to the template file.

- :meth:`~.PdsTemplate.VERSION_ID`
  Version ID of this module using two digits, e.g., "v1.0".

- :meth:`~.PdsTemplate.WRAP`
  Wrap the given text to a specified indentation and width.

These functions can also be used directly by the programmer within a template; they are
static functions of class :class:`PdsTemplate`.

##############################
Logging and Exception Handling
##############################

``pdstemplate`` employs the RMS Node's `rms-pdslogger
<https://pypi.org/project/rms-pdslogger>`_ module to handle logging. By default, the
logger is a `PdsLogger
<https://rms-pdslogger.readthedocs.io/en/latest/module.html#pdslogger.PdsLogger>`_ object,
although any ``logging.Logger`` object will work. See `pdslogger's documentation
<https://rms-pdslogger.readthedocs.io>`_ for further details.

You can override the default Logger using static method :meth:`~utils.set_logger`. You can
also set the logging level ("info", "warning", "error", etc.) using
:meth:`~utils.set_log_level` and can select among many log formatting options using
:meth:`~set_log_format`. Use :meth:`~utils.get_logger` to obtain the current Logger.

By default, exceptions during a call to :meth:`~PdsTemplate.write` or
:meth:`~PdsTemplate.generate` are handled as follows:

1. They are written to the log.
2. The expression that triggered the exception is replaced by the error text in the label,
   surrounded by "[[[" and "]]]" to make it easier to find.
3. The attributes ``fatal_count``, ``error_count``, and ``warning_count`` of the
   :class:`PdsTemplate` contain the number of messages logged by each category.
4. The exception is otherwise suppressed.

This behavior can be modified using ``raise_exceptions=True`` in the call to
:meth:`~PdsTemplate.write` or :meth:`~PdsTemplate.generate`; in this case, the exception
will be raised, label generation will stop, and the label will not be written.

##############
Pre-processors
##############

A pre-processor is a function that takes the text of a template file as input and returns
a new template as output. As described above, ``INCLUDE`` headers that contain an explicit
file name (rather than an expression to be evaluated) are handled by a pre-processor.

You may define your own functions to pre-process the content of a template. They must have
this call signature::

    func(path: str | Path | FCPath, content: str, *args, **kwargs) -> str

where

* `path` is the path to the template file (used here just for error logging).
* `content` is the content of a template represented by a single string with <LF> line
  terminators.
* `*args` is for any additional positional arguments to `func`.
* `**kwargs` is for any additional keyword arguments to `func`.

When you invoke the :meth:`PdsTemplate` constructor, one of the optional inputs is
`preprocess`, which takes either a single function or a list of functions to apply after
the INCLUDE pre-processor. For the first of these, the `args` and `kwargs` inputs can be
provided as additional inputs to the constructor. Subsequent pre-processors cannot take
additional arguments; define them using lambda notation instead.

Note that a :class:`PdsTemplate` object has an attribute `content`, which contains the
full content of the template after all pre-processing has been performed. You can examine
this attribute to see the final result of all processing. Note also that when line numbers
appear in an error message, they refer to the line number of the template after
pre-processing, not before.
"""

import datetime
import hashlib
import numbers
import os
import re
import string
import textwrap
import time
from collections import deque

from filecache import FCPath
import julian
import pdslogger

try:
    from ._version import __version__
except ImportError:                                                 # pragma: no cover
    __version__ = 'Version unspecified'

from .utils import TemplateError, TemplateAbort                     # noqa: F401
    # Unused here but included to support "from pdstemplate import TemplateError", etc.

from .utils import _RaisedException, _NOESCAPE_FLAG
from .utils import set_logger, get_logger, set_log_level, set_log_format
from ._pdsblock import _PdsBlock, _PdsIncludeBlock


class PdsTemplate:
    """Class to generate PDS labels based on a template.

    See https://rms-pdstemplate.readthedocs.io/en/latest/module.html for details.
    """

    # We need to handle certain attributes as class variables because we need to support
    # the various default functions such as LABEL_PATH(), etc., and these function execute
    # within a template without any associated context. Therefore, the PdsTemplate module
    # cannot operate in a multi-threaded environment!
    _CURRENT_TEMPLATE = None
    _CURRENT_LABEL_PATH = ''
    _CURRENT_GLOBAL_DICT = {}

    _GETENV_INCLUDE_DIRS = None

    def __init__(self, template, content='', *, xml=None, crlf=None, upper_e=False,
                 includes=[], preprocess=None, args=(), kwargs={}):
        """Construct a PdsTemplate object from the contents of a template file.

        Parameters:
            template (str, Path, or FCPath):
                Path of the input template file.
            content (str or list[str], optional):
                Alternative source of the template content rather than reading it from a
                file.
            xml (bool, optional):
                Use True to indicate that the template is in xml format; False otherwise.
                If not specified, an attempt is made to detect the format from the
                template.
            upper_e (bool, optional):
                True to force the "E" in the exponents of floating-point numbers to be
                upper case.
            crlf (bool, optional):
                True to indicate that the line termination should be <CR><LF>; False for
                <LF> only. If not specified, the line termination is inferred from the
                template.
            includes (str, Path, FCPath, or list):
                One or more directory paths where template include files can be found. The
                directory containing `template` is always searched first. Note that
                include paths can also be specified using the environment variable
                PDSTEMPLATE_INCLUDES, which should contain one or more directory paths
                separated by colons. Any directories specified here are searched before
                those defined by PDSTEMPLATE_INCLUDES.
            preprocess (function or list[function], optional):
                An optional function or list of functions that transform the content of a
                template into a new template. The call signature is
                ``func(path, content, *args, **kwargs)``
                where `path` is the path of the template file (used here just for error
                reporting), `content` is the template's content, provided as a single
                string with <LF> line terminators, and `args` and `kwargs` are additional
                inputs. Note that any pre-processor after the first cannot have `args` or
                `kwargs` as inputs; use lambda notation if later functions require inputs.
            args (tuple or list):
                Any arguments to be passed to the first preprocess function after the
                template's content.
            kwargs (dict):
                Any keywords=value arguments to be passed to the first preprocess
                function.
        """

        self.template_path = FCPath(template)
        PdsTemplate._CURRENT_TEMPLATE = self
        PdsTemplate._CURRENT_LABEL_PATH = ''
        PdsTemplate._CURRENT_GLOBAL_DICT = {}

        includes = includes if isinstance(includes, (list, tuple)) else [includes]
        self._includes = [FCPath(dir) for dir in includes]

        self.upper_e = bool(upper_e)

        logger = get_logger()
        logger.info('New PdsTemplate', self.template_path)
        try:
            # Read the template if necessary; use binary to preserve line terminators
            if not content:
                logger.debug('Reading template', self.template_path)
                content = self.template_path.read_bytes().decode('utf-8')

            # Check the line terminators
            if crlf is None:
                if isinstance(content, list):
                    crlf = content[0].endswith('\r\n')
                else:
                    crlf = content.endswith('\r\n')
                logger.debug(f'Inferred terminator is {"<CR><LF>" if crlf else "<LF>"}')
            self.crlf = crlf
            self.terminator = '\r\n' if self.crlf else '\n'

            # Convert to a single string with <LF> line terminators
            if not isinstance(content, list):
                content = content.split('\n')
                if not content[-1]:         # strip extraneous empty string at end
                    content = content[:-1]

            content = [c.rstrip('\r\n') for c in content] + ['']
            content = '\n'.join(content)

            # Preprocess the explicit $INCLUDES
            content = self._preprocess_includes(content)

            # Apply any additional preprocessor
            if preprocess:
                if not isinstance(preprocess, list):
                    preprocess = [preprocess]

                for k, func in enumerate(preprocess):
                    logger.info('Preprocessing with ' + func.__name__)
                    if k == 0:
                        content = func(self.template_path, content, *args, **kwargs)
                    else:
                        content = func(self.template_path, content)

            self.content = content

            # If the template has been pre-processed, line numbers in the error messages
            # will no longer be correct (because they are the line numbers _after_
            # pre-processing. Inside _pdsblock.py, this flag tells the logger to print the
            # actual content of the line causing the error, for simpler diagnosis of the
            # problem. DISABLED for now.
            # self._include_more_error_info = (content != before)
            self._include_more_error_info = False

            # Detect XML if not specified
            if xml is None:
                self.xml = self._detect_xml(content)
            else:
                self.xml = xml

            # Compile into a deque of _PdsBlock objects
            self._blocks = _PdsBlock.process_headers(content, self)

        except Exception as err:
            logger.exception(err, self.template_path)
            raise

        # For managing errors and warnings raised during generate()
        self.fatal_count = 0
        self.error_count = 0
        self.warning_count = 0

    def _include_dirs(self):
        """Ordered list of all include directories to search."""

        if PdsTemplate._GETENV_INCLUDE_DIRS is None:
            value = os.getenv('PDSTEMPLATE_INCLUDES')
            dirs = value.split(':') if value else []
            PdsTemplate._GETENV_INCLUDE_DIRS = [FCPath(d) for d in dirs]

        return ([self.template_path.parent] + self._includes
                + PdsTemplate._GETENV_INCLUDE_DIRS)

    _INCLUDE_REGEX = re.compile(r'(?<![^\n]) *\$INCLUDE\( *(\'[^\']+\'|"[^"]+") *\) *\n')

    def _preprocess_includes(self, content):
        """Pre-process the template content for $INCLUDE directives with explicit paths.

        Paths containing expressions of any sort are left alone.
        """

        # Split based on $INCLUDE headers. The entire template is split into substrings:
        # - Even indices contain text between the $INCLUDES
        # - Odd indices contain the file name surrounded by quotes
        parts = PdsTemplate._INCLUDE_REGEX.split(content)
        for k, part in enumerate(parts):
            if k % 2 == 1:
                part = _PdsIncludeBlock.get_content(part[1:-1], self._include_dirs())
                part = self._preprocess_includes(part)      # process recursively
                parts[k] = part

        return ''.join(parts)

    @staticmethod
    def _detect_xml(content):
        """Determine whether the given content is xml."""

        first_line = content.partition('\n')[0]

        if '<?xml' in first_line:
            return True

        count = len(first_line.split('<'))
        if count > 1 and count == len(first_line.split('>')):
            return True

        return False

    def generate(self, dictionary, label_path='', *, raise_exceptions=False,
                 hide_warnings=False, abort_on_error=False):
        """Generate the content of one label based on the template and dictionary.

        Parameters:
            dictionary (dict):
                The dictionary of parameters to replace in the template.
            label_path (str, Path, or FCPath, optional):
                The output label file path. Although a file is not written, this path is
                used in error messages.
            raise_exceptions (bool, optional):
                True to raise any exceptions encountered; False to log them and embed the
                error message into the label surrounded by "[[[" and "]]]".
            hide_warnings (bool, optional):
                True to hide warning messages.
            abort_on_error (bool, optional):
                True to abort the generation process if a validation error is encountered.
                If `raise_exceptions` is True, an exception will be raised; otherwise, the
                error will logged and an empty string will be returned.

        Returns:
            str: The generated content.
        """

        label_path = str(label_path) if label_path else ''

        # Initialize
        PdsTemplate._CURRENT_TEMPLATE = self
        PdsTemplate._CURRENT_LABEL_PATH = label_path

        # Add predefined functions to the dictionary
        global_dict = dictionary.copy()
        for name, value in PdsTemplate._PREDEFINED_FUNCTIONS.items():
            if name not in global_dict:
                global_dict[name] = value
        global_dict['hide_warnings'] = bool(hide_warnings)
        global_dict['abort_on_error'] = bool(abort_on_error)
        PdsTemplate._CURRENT_GLOBAL_DICT = global_dict

        state = _LabelState(self, global_dict, label_path,
                            raise_exceptions=raise_exceptions)

        # Generate the label content recursively
        results = deque()
        logger = get_logger()
        logger.open('Generating label', label_path)
        try:
            for block in self._blocks:
                results += block.execute(state)
        except TemplateAbort as err:
            logger.fatal('**** ' + err.message, label_path)
        finally:
            (fatals, errors, warns, total) = logger.close()

        content = ''.join(results)
        self.fatal_count = fatals
        self.error_count = errors
        self.warning_count = warns

        # Update the terminator if necessary
        if self.terminator != '\n':
            content = content.replace('\n', self.terminator)

        # Reset global symbols
        PdsTemplate._CURRENT_LABEL_PATH = ''
        PdsTemplate._CURRENT_GLOBAL_DICT = {}

        return content

    def write(self, dictionary, label_path, *, mode='save', backup=False,
              raise_exceptions=False, handler=None):
        """Write one label based on the template, dictionary, and output filename.

        Parameters:
            dictionary (dict):
                The dictionary of parameters to replace in the template.
            label_path (str, Path, or FCPath, optional):
                The output label file path.
            mode (str, optional):
                "save" to save the new label content regardless of any warnings or errors;
                "repair" to save the new label if warnings occurred but no errors;
                "validate" to log errors and warnings but never save the new label file.
            backup (bool, optional):
                If True and an existing file of the same name as label_path already
                exists, that file is renamed with a suffix indicating its original
                modification date. The format is "_yyyy-mm-ddThh-mm-ss" and it appears
                after the file stem, before the extension.
            raise_exceptions (bool, optional):
                True to raise any exceptions encountered; False to log them and embed the
                error message into the label surrounded by "[[[" and "]]]".
            handler (str, Path, FCPath, or logger.Handler, optional):
                Define a handler to use exclusively during the generation of this label.
                If it is a string, Path, or FCPath, a logger.FileHandler using this path
                is constructed. However, if the string begins with "." is instead
                interpreted as the file extension to use, where the path up to the
                extension is defined by `label_path`; for example, if handler=".log", then
                when writing "path/to/123.lbl", the log created will be "path/to/123.log".
                This log is automatically closed once the label is written.

        Returns:
            int: Number of errors issued.
            int: Number of warnings issued.
        """

        if mode not in {'save', 'repair', 'validate'}:
            raise ValueError('invalid mode value: ' + repr(mode))

        label_path = FCPath(label_path)

        logger = get_logger()
        if handler:
            if isinstance(handler, str):
                if handler.startswith('.'):
                    handler = label_path.with_suffix(handler)
                else:
                    handler = FCPath(handler)
            if isinstance(handler, FCPath):
                handler = pdslogger.file_handler(handler)
            logger.add_handler(handler)

        try:
            content = self.generate(dictionary, label_path,
                                    raise_exceptions=raise_exceptions,
                                    hide_warnings=(mode == 'save'),
                                    abort_on_error=(mode != 'save'))
            fatals = self.fatal_count
            errors = self.error_count
            warns = self.warning_count

            if fatals and not errors:
                errors = fatals

            # Validation case
            if mode == 'validate':
                if errors:
                    plural = 's' if errors > 1 else ''
                    logger.error(f'Validation failed with {errors} error{plural}',
                                 label_path, force=True)
                elif warns:
                    plural = 's' if warns > 1 else ''
                    logger.warning(f'Validation failed with {warns} warning{plural}',
                                   label_path, force=True)
                else:
                    logger.info('Validation successful', label_path, force=True)

            # Repair case
            elif mode == 'repair':
                if errors:
                    plural = 's' if errors > 1 else ''
                    logger.warning(f'Repair failed with {errors} error{plural}',
                                   label_path)
                elif label_path.exists():
                    old_content = label_path.read_bytes().decode('utf-8')
                    if old_content == content:
                        logger.info('Repair unnecessary; content is unchanged',
                                    label_path)
                    else:
                        mode = 'save'       # re-save the file
                else:
                    plural = 's' if warns > 1 else ''
                    logger.info(f'Repairing {warns} warning{plural}', label_path,
                                force=True)
                    mode = 'save'           # proceed with saving the file

            # Otherwise, save
            if mode != 'save':
                return (errors, warns)

            # Don't save a file after a fatal error
            if fatals:
                logger.error('File save aborted due to prior errors')
                return (errors, warns)

            # Backup existing label if necessary
            exists = label_path.exists()
            if exists and backup:
                timestamp = os.path.getmtime(label_path.get_local_path())
                date = datetime.datetime.fromtimestamp(timestamp)   # wrong if remote
                datestr = date.isoformat(timespec='seconds').replace(':', '-')
                backup_path = label_path.parent / (label_path.stem + '_' + datestr +
                                                   label_path.suffix)
                label_path.rename(backup_path)
                logger.info('Existing label renamed to', backup_path)
                exists = False

            # Write label
            if content and not content.endswith(self.terminator):
                content += self.terminator
            label_path.write_bytes(content.encode('utf-8'))

            # Log event
            if exists:
                logger.info('Label re-written', label_path)
            else:
                logger.info('Label written', label_path)

        finally:
            logger.remove_handler(handler)      # OK if handler is None

        return (errors, warns)

    @staticmethod
    def log(level, message, filepath='', *, force=False):
        """Send a message to the current logger.

        This allows external modules to issue warnings and other messages.

        Parameters:
            level (int or str): Level of the message: 'info', 'error', 'warn', etc.
            message (str): Text of the warning message.
            filepath (str, Path, or FCPath, optional): File path to include in the
                message.
            force (bool, optional): True to force the logging of the message regardless
                of the level.
        """

        get_logger().log(level, message, filepath, force=force)

    @staticmethod
    def define_global(name, value):
        """Define a new global symbol.

        This allows external modules to define new symbols during template generation.

        Parameters:
            name (str): Name of global symbol as it will appear inside the template.
            value (any): Value of the symbol.
        """

        # Add the new value to the permanent set (even if it's not really a function)
        PdsTemplate._PREDEFINED_FUNCTIONS[name] = value

        # If generate() is currently active, add it to the active dictionary too
        if PdsTemplate._CURRENT_LABEL_PATH:     # hard to get here  # pragma: no cover
            PdsTemplate._CURRENT_GLOBAL_DICT[name] = value

    ######################################################################################
    # Utility functions
    ######################################################################################

    @staticmethod
    def BASENAME(filepath):
        """The basename of `filepath`, with the leading directory path removed.

        Parameters:
            filepath (str): The filepath.

        Returns:
            str: The basename of the filepath (the final filename).
        """

        return os.path.basename(filepath)

    @staticmethod
    def BOOL(value, true='true', false='false'):
        """Return `true` if `value` evaluates to Boolean True; otherwise, return `false`.

        Parameters:
            value (truthy): The expression to evaluate for truthy-ness.
            true (str, optional): The value to return for a True expression.
            false (str, optional): The value to return for a False expression.

        Returns:
            str: "true" or "false", or the given values in the `true` and/or `false`
            parameters.
        """

        return (true if value else false)

    _counters = {}

    @staticmethod
    def COUNTER(name, reset=False):
        """The value of a counter identified by `name`, starting at 1.

        Parameters:
            name (str): The name of the counter. If the counter has not been used
                before, it will start with a value of 1.
            reset (bool, optional): If True, reset the counter to a value of zero
                and return the value 0. The next time this counter is referenced,
                it will have the value 1.

        Returns:
            int: The value of the counter.
        """

        if name not in PdsTemplate._counters.keys():
            PdsTemplate._counters[name] = 0
        PdsTemplate._counters[name] += 1
        if reset:
            PdsTemplate._counters[name] = 0
        return PdsTemplate._counters[name]

    @staticmethod
    def CURRENT_TIME(date_only=False):
        """The current date/time in the local time zone.

        Parameters:
            date_only (bool, optional): Return only the date without the time.

        Returns:
            str: The current date/time in the local time zone as a formatted string of
            the form "yyyy-mm-ddThh:mm:sss" if `date_only=False` or "yyyy-mm-dd" if
            `date_only=True`.
        """

        if date_only:
            return datetime.datetime.now().isoformat()[:10]
        return datetime.datetime.now().isoformat()[:19]

    @staticmethod
    def CURRENT_ZULU(date_only=False):
        """The current UTC date/time.

        Parameters:
            date_only (bool, optional): Return only the date without the time.

        Returns:
            str: The current date/time in UTC as a formatted string of the form
            "yyyy-mm-ddThh:mm:sssZ" if `date_only=False` or "yyyy-mm-dd" if
            `date_only=True`.
        """

        if date_only:
            return time.strftime('%Y-%m-%d', time.gmtime())
        return time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())

    @staticmethod
    def _DATETIME(value, offset=0, digits=None, date_type='YMD'):
        """Convert the given date/time string or time in TDB seconds to a year-month-day
        format with a trailing "Z". The date can be in any format parsable by the Julian
        module. An optional offset in seconds is applied. If the value is "UNK", then
        "UNK" is returned.
        """

        if isinstance(value, numbers.Real):
            if digits is None:
                digits = 3

            tai = julian.tai_from_tdb(value)

            # Convert to ISO format or return seconds
            if date_type in ('YMDT', 'YDT'):
                return julian.format_tai(tai + offset, order=date_type, sep='T',
                                         digits=digits, suffix='Z')
            else:
                (day, sec) = julian.day_sec_from_tai(tai + offset)
                return sec

        if value.strip() == 'UNK':
            return 'UNK'

        # Convert to day and seconds
        (day, sec) = julian.day_sec_from_string(value, timesys=True)[:2]

        # Retain the number of digits precision in the source, if appropriate
        if digits is None and offset % 1 == 0:
            parts = re.split(r'\d\d:\d\d:\d\d', value)
            if len(parts) == 2 and parts[1].startswith('.'):
                digits = len(re.match(r'(\.\d*)', parts[1]).group(1)) - 1

        # Apply offset if necessary
        if offset:
            tai = julian.tai_from_day_sec(day, sec)
            (day, sec) = julian.day_sec_from_tai(tai + offset)

        # Interpret the number of digits if still unknown
        if digits is None:
            if sec % 1 == 0.:
                digits = -1     # no fractional part, no decimal point
            else:
                digits = 3
        elif digits == 0:
            digits = -1         # suppress decimal point

        # Convert to ISO format or return seconds
        if date_type in ('YMDT', 'YDT'):
            return julian.format_day_sec(day, sec, order=date_type, sep='T',
                                         digits=digits, suffix='Z')
        else:
            return sec

    @staticmethod
    def DATETIME(time, offset=0, digits=None):
        """Convert `time` to an ISO date of the form "yyyy-mm-ddThh:mm:ss[.fff]Z".

        Parameters:
            time (str or float): The time as an arbitrary date/time string or TDB seconds.
                If `time` is "UNK", then "UNK" is returned.
            offset (float, optional): The offset, in seconds, to add to the time.
            digits (int, optional): The number of digits after the decimal point in the
                seconds field to return. If not specified, the appropriate number of
                digits for the time is used.

        Returns:
            str: The time in the format "yyyy-mm-ddThh:mm:ss[.fff]Z".
        """

        return PdsTemplate._DATETIME(time, offset, digits, date_type='YMDT')

    @staticmethod
    def DATETIME_DOY(time, offset=0, digits=None):
        """Convert `time` to an ISO date of the form "yyyy-dddThh:mm:ss[.fff]Z".

        Parameters:
            time (str or float): The time as an arbitrary date/time string or TDB seconds.
                If `time` is "UNK", then "UNK" is returned.
            offset (float, optional): The offset, in seconds, to add to the time.
            digits (int, optional): The number of digits after the decimal point in the
                seconds field to return. If not specified, the appropriate number of
                digits for the time is used.

        Returns:
            str: The time in the format "yyyy-dddThh:mm:ss[.fff]Z".
        """

        return PdsTemplate._DATETIME(time, offset, digits, date_type='YDT')

    @staticmethod
    def DAYSECS(time):
        """The number of elapsed seconds since the most recent midnight.

        Parameters:
            time (str or float): The time as an arbitrary date/time string or TDB seconds.
                If `time` is "UNK", then "UNK" is returned.

        Returns:
            float: The number of elapsed seconds since the most recent midnight.
        """

        if isinstance(time, numbers.Real):
            return PdsTemplate._DATETIME(time, 0, None, date_type='SEC')

        try:
            return julian.sec_from_string(time)
        except Exception:
            return PdsTemplate._DATETIME(time, 0, None, date_type='SEC')

    @staticmethod
    def FILE_BYTES(filepath):
        """The size in bytes of the file specified by `filepath`.

        Parameters:
            filepath (str): The filepath.

        Returns:
            int: The size in bytes of the file.
        """

        return os.path.getsize(filepath)

    # From http://stackoverflow.com/questions/3431825/-
    @staticmethod
    def FILE_MD5(filepath):
        """The MD5 checksum of the file specified by `filepath`.

        Parameters:
            filepath (str): The filepath.

        Returns:
            str: The MD5 checksum of the file.
        """

        blocksize = 65536
        with open(filepath, 'rb') as f:
            hasher = hashlib.md5()
            buf = f.read(blocksize)
            while len(buf) > 0:
                hasher.update(buf)
                buf = f.read(blocksize)

        f.close()
        return hasher.hexdigest()

    @staticmethod
    def FILE_RECORDS(filepath):
        """The number of records in the the file specified by `filepath`.

        Parameters:
            filepath (str): The filepath.

        Returns:
            int: The number of records in the file if it is ASCII;
            0 if the file is binary.
        """

        # We intentionally open this in non-binary mode so we don't have to
        # content with line terminator issues
        with open(filepath, 'r') as f:
            count = 0
            asciis = 0
            non_asciis = 0
            for line in f:
                for c in line:
                    if c in string.printable:
                        asciis += 1
                    else:
                        non_asciis += 1

                count += 1

        if non_asciis > 0.05 * asciis:
            return 0

        return count

    @staticmethod
    def FILE_TIME(filepath):
        """The modification time in the local time zone of a file.

        Parameters:
            filepath (str): The filepath.

        Returns:
            str: The modification time in the local time zone of the file specified by
            `filepath` in the form "yyyy-mm-ddThh:mm:ss".
        """

        timestamp = os.path.getmtime(filepath)
        return datetime.datetime.fromtimestamp(timestamp).isoformat()[:19]

    @staticmethod
    def FILE_ZULU(filepath):
        """The UTC modification time of a file.

        Parameters:
            filepath (str): The filepath.

        Returns:
            str: The UTC modification time of the file specified by `filepath` in the
            form "yyyy-mm-ddThh:mm:ssZ".
        """

        timestamp = os.path.getmtime(filepath)
        try:
            utc_dt = datetime.datetime.fromtimestamp(timestamp, datetime.UTC)
        except AttributeError:  # pragma: no cover
            # Python < 3.11
            utc_dt = datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)
        return utc_dt.strftime('%Y-%m-%dT%H:%M:%SZ')

    @staticmethod
    def GETENV(name, default=''):
        """The value of the specified environment variable.

        Parameters:
            name (str): Name of the environment variable.
            default (str): Value to return if the environment variable is undefined.

        Returns:
            str: The value of the variable or else the default.
        """

        return os.getenv(name, default=default)

    @staticmethod
    def LABEL_PATH():
        """The path to the current label file being generated.

        Returns:
            str: Path string to the label file.
        """

        return str(PdsTemplate._CURRENT_LABEL_PATH)

    @staticmethod
    def LOG(level, message, filepath='', *, force=False):
        """Send a message to the logger; nothing is returned.

        This allows a template to issue warnings and other messages.

        Parameters:
            level (int or str): Level of the message: 'info', 'error', 'warn', etc.
            message (str): Text of the warning message.
            filepath (str, Path, or FCPath, optional): File path to include in the
                message.
            force (bool, optional): True to force the logging of the message regardless
                of the level.
        """

        get_logger().log(level, message, filepath, force=force)

    @staticmethod
    def NOESCAPE(text):
        """Prevent the given text from being escaped in the XML.

        If the template is XML, evaluated expressions are "escaped" to ensure that they
        are suitable for embedding in a PDS label. For example, ">" inside a string will
        be replaced by "&gt;". This function prevents `text` from being escaped in the
        label, allowing it to contain literal XML.

        Parameters:
            text (str): The text that should not be escaped.

        Returns:
            str: The text marked so that it won't be escaped.
        """

        return _NOESCAPE_FLAG + text

    @staticmethod
    def QUOTE_IF(text):
        """Place the given text in quotes if it contains any character other than letters,
        digits, and an underscore.

        An empty string is also quoted. A string that is already enclosed in quotes is
        not quoted (but quote balancing is not checked). Other values are returned
        unchanged.

        Parameters:
            text (str): Text to possibly quote.

        Returns:
            str: The text with quotes if necessary.
        """

        if text.isidentifier():
            return text

        if text.startswith('"') and text.endswith('"'):
            return text

        return '"' + text + '"'

    @staticmethod
    def RAISE(exception, message):
        """Raise an exception with the given class `exception` and the `message`.

        Parameters:
            exception (type): The class of the exception to raise, e.g., ValueError.
            message (str): The message to include in the exception.

        Raises:
            Exception: The specified exception.
        """

        raise _RaisedException(exception, message)  # wrapper used to handle formatting

    @staticmethod
    def REPLACE_NA(value, na_value, flag='N/A'):
        """Return `na_value` if `value` equals "N/A"; otherwise, return `value`.

        Parameters:
            value (str or int or float or bool): The input value.
            flag (str or int or float or bool, optional): The value that means N/A.
                Defaults to the string "N/A".

        Returns:
            str or int or float or bool: The original value if it is not equal to
            `flag`, otherwise `na_value`.
        """

        if isinstance(value, str):
            value = value.strip()

        if value == flag:
            return na_value
        else:
            return value

    @staticmethod
    def REPLACE_UNK(value, unk_value):
        """Return `unk_value` if `value` equals "UNK"; otherwise, return `value`.

        Parameters:
            value (str or int or float or bool): The input value.

        Returns:
            str or int or float or bool: The original value if it is not equal to
            "UNK", otherwise `unk_value`.
        """

        return PdsTemplate.REPLACE_NA(value, unk_value, flag='UNK')

    @staticmethod
    def TEMPLATE_PATH():
        """The path to this template file.

        Returns:
            str: Path string to this template file.
        """

        return str(PdsTemplate._CURRENT_TEMPLATE.template_path)

    @staticmethod
    def VERSION_ID():
        """The PdsTemplate version ID using two digits, e.g., "v1.0".

        Returns:
            str: The version ID.
        """

        parts = __version__.split('.')
        if len(parts) >= 2:                                         # pragma: no cover
            return '.'.join(parts[:2])

        return '0.0'        # version unspecified                   # pragma: no cover

    @staticmethod
    def WRAP(left, right, text, preserve_single_newlines=True):
        """Format `text` to fit between the `left` and `right` column numbers.

        The first line is not indented, so the text will begin in the column where "$WRAP"
        first appears in the template.

        Parameters:
            left (int): The starting column number, numbered from 0.
            right (int): the ending column number, numbered from 0.
            text (str): The text to wrap.
            preserve_single_newlines (bool, optional): If True, single newlines
                are preserved. If False, single newlines are just considered to be
                wrapped text and do not cause a break in the flow.

        Returns:
            str: The wrapped text.
        """

        if not preserve_single_newlines:
            # Remove any newlines between otherwise good text - we do this twice
            #   because sub is non-overlapping and single-character lines won't
            #   get treated properly
            # Remove any single newlines at the beginning or end of the string
            # Remove any pair of newlines after otherwise good text
            # Remove any leading or trailing spaces
            text = re.sub(r'([^\n])\n([^\n])', r'\1 \2', text)
            text = re.sub(r'([^\n])\n([^\n])', r'\1 \2', text)
            text = re.sub(r'([^\n])\n$', r'\1', text)
            text = re.sub(r'^\n([^\n])', r'\1', text)
            text = re.sub(r'([^\n])\n\n', r'\1\n', text)
            text = text.strip(' ')

        old_lines = text.splitlines()

        indent = left * ' '
        new_lines = []
        for line in old_lines:
            if line:
                new_lines += textwrap.wrap(line,
                                           width=right,
                                           initial_indent=indent,
                                           subsequent_indent=indent,
                                           break_long_words=False,
                                           break_on_hyphens=False)
            else:
                new_lines.append('')

        # strip the first left indent; this should be where "$WRAP" appears in the
        # template.
        new_lines[0] = new_lines[0][left:]

        return '\n'.join(new_lines)


PdsTemplate._PREDEFINED_FUNCTIONS = {}
PdsTemplate._PREDEFINED_FUNCTIONS['BASENAME'     ] = PdsTemplate.BASENAME
PdsTemplate._PREDEFINED_FUNCTIONS['BOOL'         ] = PdsTemplate.BOOL
PdsTemplate._PREDEFINED_FUNCTIONS['COUNTER'      ] = PdsTemplate.COUNTER
PdsTemplate._PREDEFINED_FUNCTIONS['CURRENT_TIME' ] = PdsTemplate.CURRENT_TIME
PdsTemplate._PREDEFINED_FUNCTIONS['CURRENT_ZULU' ] = PdsTemplate.CURRENT_ZULU
PdsTemplate._PREDEFINED_FUNCTIONS['DATETIME'     ] = PdsTemplate.DATETIME
PdsTemplate._PREDEFINED_FUNCTIONS['DATETIME_DOY' ] = PdsTemplate.DATETIME_DOY
PdsTemplate._PREDEFINED_FUNCTIONS['DAYSECS'      ] = PdsTemplate.DAYSECS
PdsTemplate._PREDEFINED_FUNCTIONS['FILE_BYTES'   ] = PdsTemplate.FILE_BYTES
PdsTemplate._PREDEFINED_FUNCTIONS['FILE_MD5'     ] = PdsTemplate.FILE_MD5
PdsTemplate._PREDEFINED_FUNCTIONS['FILE_RECORDS' ] = PdsTemplate.FILE_RECORDS
PdsTemplate._PREDEFINED_FUNCTIONS['FILE_TIME'    ] = PdsTemplate.FILE_TIME
PdsTemplate._PREDEFINED_FUNCTIONS['FILE_ZULU'    ] = PdsTemplate.FILE_ZULU
PdsTemplate._PREDEFINED_FUNCTIONS['GETENV'       ] = PdsTemplate.GETENV
PdsTemplate._PREDEFINED_FUNCTIONS['LABEL_PATH'   ] = PdsTemplate.LABEL_PATH
PdsTemplate._PREDEFINED_FUNCTIONS['LOG'          ] = PdsTemplate.LOG
PdsTemplate._PREDEFINED_FUNCTIONS['NOESCAPE'     ] = PdsTemplate.NOESCAPE
PdsTemplate._PREDEFINED_FUNCTIONS['QUOTE_IF'     ] = PdsTemplate.QUOTE_IF
PdsTemplate._PREDEFINED_FUNCTIONS['RAISE'        ] = PdsTemplate.RAISE
PdsTemplate._PREDEFINED_FUNCTIONS['REPLACE_NA'   ] = PdsTemplate.REPLACE_NA
PdsTemplate._PREDEFINED_FUNCTIONS['REPLACE_UNK'  ] = PdsTemplate.REPLACE_UNK
PdsTemplate._PREDEFINED_FUNCTIONS['TEMPLATE_PATH'] = PdsTemplate.TEMPLATE_PATH
PdsTemplate._PREDEFINED_FUNCTIONS['VERSION_ID'   ] = PdsTemplate.VERSION_ID
PdsTemplate._PREDEFINED_FUNCTIONS['WRAP'         ] = PdsTemplate.WRAP

##########################################################################################
# LabelStatus class
##########################################################################################

class _LabelState(object):
    """Internal class to carry status information about where we are in the template and
    the label generation.

    Parameters:
        template (PdsTemplate): The template being processed into a label file.
        dictionary (dict): The dictionary of values used for substitutions.
        label_path (str, path, or FCPath): The path to the file being generated.
        terminator (str, optional):
            The line terminator, either "\\n" or "\\r\\n". The default is to retain the
            line terminator used in the template.
        raise_exceptions (bool, optional):
            True to raise any exceptions encountered; False to log them and embed the
            error messages into the label, marked by "[[[" and "]]]".
    """

    def __init__(self, template, dictionary, label_path='', *, terminator=None,
                 raise_exceptions=False):

        self.template = template
        self.label_path = label_path
        self.terminator = terminator
        self.raise_exceptions = raise_exceptions

        self.local_dicts = [{}]

        # Merge the predefined functions into a copy of the global dictionary
        self.global_dict = dictionary.copy()
        for key, func in PdsTemplate._PREDEFINED_FUNCTIONS.items():
            if key not in self.global_dict:
                self.global_dict[key] = func

    def define_global(self, name, value):
        """Add this definition to this state's global dictionary."""

        self.global_dict[name] = value

##########################################################################################
# Allow access of key functions a static methods of PdsTemplate
##########################################################################################

PdsTemplate.set_logger = set_logger
PdsTemplate.get_logger = get_logger
PdsTemplate.set_log_level = set_log_level
PdsTemplate.set_log_format = set_log_format

##########################################################################################



##########################################################################################
# pdstemplate/utils.py
##########################################################################################
"""
#################
pdstemplate.utils
#################

Utility functions and classes.
"""

from filecache import FCPath
from pdslogger import PdsLogger, LoggerError


class TemplateError(LoggerError):
    """Class for all template parsing exceptions."""
    pass


class TemplateAbort(TemplateError):
    """Raise this class to abort processing a template if the situation is hopeless.

    When raising this exception, pass it the exact text to appear in the log. It will
    appear as a FATAL message with four asterisks followed by the given text, and the file
    path if any.
    """
    pass


class _RaisedException(Exception):
    """Internal; used to manage error messages from `RAISE()`."""

    def __init__(self, exception, message):
        self.exception = exception
        self.message = message


_NOESCAPE_FLAG = '!!NOESCAPE!!:'    # used internally

##########################################################################################
# Logger management
##########################################################################################

# Define the global logger with streamlined output, no handlers so printing to stdout
_LOGGER = PdsLogger.get_logger('pds.template', timestamps=False, digits=0, lognames=False,
                               pid=False, indent=True, blanklines=False, level='info')

def set_logger(logger):
    """Define the global logger for PdsTemplate and associated tools.

    Parameters:
        logger (PdsLogger or logging.Logger): Logger to use, replacing the default.

    Returns:
        PdsLogger: The new PdsLogger. If the input was a logging.Logger, it is converted
        to a PdsLogger.
    """

    global _LOGGER

    _LOGGER = PdsLogger.as_pdslogger(logger)
    return _LOGGER


def get_logger():
    """The global PdsLogger for PdsTemplate and associated tools."""

    return _LOGGER


def set_log_level(level):
    """Set the minimum level for messages to be logged.

    Parameters:
        level (int or str, optional):
            The minimum level of level name for a record to enter the log. Use an integer
            1-50 or a level name, one of "debug"=10, "info"=20, "warning"=30, "error"=40,
            or "fatal"=50.
    """

    _LOGGER.set_level(level)


def set_log_format(**kwargs):
    """Set the formatting and other properties of the logger.

    Parameters:
        level (int or str, optional):
            The minimum level of level name for a record to enter the log.
        timestamps (bool, optional):
            True or False, defining whether to include a timestamp in each log record.
        digits (int, optional):
            Number of fractional digits in the seconds field of the timestamp.
        lognames (bool, optional):
            True or False, defining whether to include the name of the logger in each log
            record.
        pid (bool, optional):
            True or False, defining whether to include the process ID in each log record.
        indent (bool, optional):
            True or False, defining whether to include a sequence of dashes in each log
            record to provide a visual indication of the tier in a logging hierarchy.
        blanklines (bool, optional):
            True or False, defining whether to include a blank line in log files when a
            tier in the hierarchy is closed.
        colors (bool, optional):
            True or False, defining whether to color-code the log files generated, for
            Macintosh only.
        maxdepth (int, optional):
            Maximum depth of the logging hierarchy, needed to prevent unlimited recursion.
    """

    _LOGGER.set_format(**kwargs)

##########################################################################################
# Line terminator utility
##########################################################################################

def _check_terminators(filepath, content='', crlf=None):
    """Raise an exception if the given file content is not consistent with the intended
    line terminator.

    Parameters:
        filepath (str, Path, or FCPath):
            Path to the file, used for error messages.
        content (str, bytes, list[str], or list[bytes]):
            Content of the file as a single string or byte string or else as a list of
            records such as provided by file.readlines().
        crlf (bool, optional):
            True to indicate that the line termination should be <CR><LF>; False for
            <LF> only. If not specified, the line termination is inferred from the
            template.

    Returns:
        bool: True if the terminators are <CR><LF>, False otherwise.

    Raises:
        TemplateError: If an incorrect line terminator was found.
    """

    filepath = FCPath(filepath)
    if not content:
        content = filepath.read_bytes()

    # Define <CR><LF> terminator depending on types; split content into a list
    if isinstance(content, list):
        crlf_chars = b'\r\n' if isinstance(content[0], bytes) else '\r\n'
    else:
        lf_char = b'\n' if isinstance(content, bytes) else '\n'
        content = content.split(lf_char)
        if content[-1]:
            raise TemplateError('missing line terminator at end of file', filepath)
        content = content[:-1]
        crlf_chars = b'\r' if isinstance(content[0], bytes) else '\r'
            # Because the split was on the <LF>, we do not expect <LF>'s in records

    # Define `crlf` if it was not provided
    if crlf is None:
        crlf = content[0].endswith(crlf_chars)

    # Validate line terminator in first record
    crlf = bool(crlf)   # make sure it's really boolean
    if crlf != content[0].endswith(crlf_chars):
        name = '<CR><LF>' if crlf else '<LF>'
        raise TemplateError(f'Line terminator is not {name}', filepath)

    # Validate the line terminator in every record
    for recno, record in enumerate(content):
        if crlf != record.endswith(crlf_chars):
            raise TemplateError(f'Inconsistent line terminator at line {recno+1}',
                                filepath)

    return crlf

##########################################################################################



##########################################################################################
# tests/test_utils.py
##########################################################################################

import unittest

from pdstemplate import TemplateError
from pdstemplate.utils import _check_terminators


class Test_Utils(unittest.TestCase):

    def test_check_terminators(self):

        text = 'line 1\nline 2\nline 3\n'
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), False)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=False), False)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=True)

        text = 'line 1\r\nline 2\r\nline 3\r\n'
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), True)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=True), True)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=False)

        text = ['line 1\n', 'line 2\n', 'line 3\n']
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), False)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=False), False)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=True)

        text = ['line 1\r\n', 'line 2\r\n', 'line 3\r\n']
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), True)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=True), True)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=False)

        text = b'line 1\nline 2\nline 3\n'
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), False)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=False), False)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=True)

        text = b'line 1\r\nline 2\r\nline 3\r\n'
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), True)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=True), True)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=False)

        text = [b'line 1\n', b'line 2\n', b'line 3\n']
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), False)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=False), False)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=True)

        text = [b'line 1\r\n', b'line 2\r\n', b'line 3\r\n']
        self.assertEqual(_check_terminators('foo.txt', text, crlf=None), True)
        self.assertEqual(_check_terminators('foo.txt', text, crlf=True), True)
        self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text, crlf=False)

        # Missing final terminator
        for text in ('line 1\nline 2\nline 3', b'line 1\nline 2\nline 3',
                     'line 1\r\nline 2\r\nline 3', b'line 1\r\nline 2\r\nline 3'):
            for crlf in (False, True, None):
                self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text,
                                  crlf=crlf)

        # Mixed terminators
        for text in ('line 1\nline 2\nline 3\r\n', b'line 1\nline 2\nline 3\r\n',
                     'line 1\r\nline 2\r\nline 3\n', b'line 1\r\nline 2\n\nline 3\n'):
            for crlf in (False, True, None):
                self.assertRaises(TemplateError, _check_terminators, 'foo.txt', text,
                                  crlf=crlf)



#!/usr/bin/env python3
##########################################################################################
# rms-pdstemplate/programs/tablelabel.py
##########################################################################################
"""
.. tablelabel:

##########
tablelabel
##########

This is a stand-alone program that can be used to validate and repair existing PDS3 labels
describing ASCII tables and can also create new labels. Type::

    tablelabel --help

for more information.
"""

import argparse
import pathlib
import sys

import pdslogger
from pdstemplate import PdsTemplate, set_logger
from pdstemplate.pds3table import pds3_table_preprocessor

LOGNAME = 'pds.tablelabel'

# Set up parser
parser = argparse.ArgumentParser(
    description='tablelabel: Validate, repair, or create a PDS3 label file for an '
                'existing ASCII table.')

group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--validate', '-v',
                   dest='task', action='store_const', const='validate',
                   help='Validate an existing label, logging any errors or other '
                        'discrepancies as warnings messages; do not write a new label.')

group.add_argument('--repair', '-r',
                   dest='task', action='store_const', const='repair',
                   help='Update an existing label file only if the new label would be '
                        'different; otherwise leave it unchanged.')

group.add_argument('--create', '-c',
                   dest='task', action='store_const', const='create',
                   help='Create a new label where none exists; leave existing labels '
                        'alone.')

group.add_argument('--save', '-s',
                   dest='task', action='store_const', const='save',
                   help='Save a new label, replacing any existing file.')

parser.add_argument('path', nargs='*', type=str,
                    help='Path to one or more PDS3 label or ASCII table files. It is '
                         'always assumed that a label file has a ".lbl" extension and '
                         'its associated table file is in the same directory but with a '
                         '".tab" extension.')

parser.add_argument('--template', '-t', type=str,
                    help='An optional template file path. If specified, this template is '
                         'used to generate new label content; otherwise, an existing '
                         'label is validated or repaired.')

parser.add_argument('--numbers', '-n', action='store_true',
                    help='Require every column to have a COLUMN_NUMBER attribute.')

parser.add_argument('--formats', '-f', action='store_true',
                    help='Require every column to have a FORMAT attribute.')

parser.add_argument('--minmax', '-m', nargs='*', type=str,
                    help='One or more column names that should have the MINIMUM_VALUE '
                         'and MAXIMUM_VALUE attributes. Use "float" to include these '
                         'attributes for all floating-point columns; use "int" to '
                         'include these attributes for all integer-valued columns.')

parser.add_argument('--derived', nargs='*', type=str,
                    help='One or more column names that should have the DERIVED_MINIMUM '
                         'and DERIVED_MAXIMUM attributes. Use "float" to include these '
                         'attributes for all floating-point columns.')

parser.add_argument('--edit', nargs='*', type=str,
                    help='One or more expressions of the form "column:name=value", '
                         'which will be used to insert or replace values currently in '
                         'the label.')

parser.add_argument('--real', nargs='*', type=str,
                    help='One or more COLUMN names that should be identified as '
                         'ASCII_REAL even if every value in the table is an integer.')

parser.add_argument('--dict', '-d', nargs='*', type=str,
                    help='One or more keyword definitions of the form "name=value", '
                         'which will be used when the label is generated. Each value '
                         'must be an integer, float, or quoted string.')

parser.add_argument('-e', dest='upper_e', action='store_false', default=True,
                    help='Format values involving an exponential using lower case "e"')

parser.add_argument('-E', dest='upper_e', action='store_true',
                    help='Format values involving an exponential using upper case "E"')

parser.add_argument('--nobackup', '-B', action='store_true',
                    help='If a file is repaired, do not save a backup of an existing '
                         'file. Otherwise, an existing file is renamed with a suffix '
                         'identifying its original creation date and time.')

parser.add_argument('--quiet', '-q', action='store_true',
                    help='Do not log to the terminal.')

parser.add_argument('--log', '-l', action='store_true',
                    help='Save a log file of warnings and steps performed. The log file '
                         'will have the same name as the label except the extension '
                         'will be ".log" instead of ".lbl".')

parser.add_argument('--debug', action='store_true',
                    help='Include "debug" messages in the log.')

parser.add_argument('--timestamps', action='store_true',
                    help='Include a timestamp in each log record.')

# Parse and validate the command line
args = parser.parse_args()

kwargs = {
    'numbers' : args.numbers,
    'formats' : args.formats,
    'minmax'  : args.minmax or [],
    'derived' : args.derived or [],
    'edits'   : args.edit or [],
    'reals'   : args.real or [],
    'validate': True,
}

# Interpret paths
paths = [pathlib.Path(p).with_suffix('.lbl') for p in args.path]

# Define the logger
logger = pdslogger.PdsLogger.get_logger(LOGNAME, timestamps=args.timestamps, digits=3,
                                        lognames=False, indent=True, blanklines=False,
                                        level='debug' if args.debug else 'info')
logger.add_handler(pdslogger.NULL_HANDLER)

parents = set(p.parent for p in paths)
logger.add_root(*parents)

# Define the default handlers
set_logger(logger)
if not args.quiet:
    logger.add_handler(pdslogger.STDOUT_HANDLER)

# Intepret the template if provided
template = None
if args.template:
    template = PdsTemplate(args.template, crlf=True, upper_e=args.upper_e,
                           preprocess=pds3_table_preprocessor, kwargs=kwargs)
    if len(paths) > 1:
        logger.blankline()

# Interpret the --dict input
dictionary = {}
for item in args.dict or []:
    name, _, value = item.partition('=')
    try:
        value = eval(value)
    except Exception:
        pass
    dictionary[name] = value

# Process each path...
errors = 0
warnings = 0
for k, path in enumerate(paths):

    # Skip existing files for task == "create"
    if args.task == 'create' and path.exists():
        logger.info('Existing file skipped', path)
        if k > len(paths) - 1:
            logger.blankline()
        continue

    # If there's not a default template, each label is its own template
    if not args.template:
        template = PdsTemplate(path, crlf=True, upper_e=args.upper_e,
                               preprocess=pds3_table_preprocessor, kwargs=kwargs)

    # Save a log file for each path if necessary
    handler = None
    if args.log:
        handler = pdslogger.file_handler(path.with_suffix('.log'))
        logger.add_handler(handler)

    # Process one label
    mode = 'save' if args.task == 'create' else args.task
    try:
        status = template.write(dictionary, path, mode=mode, backup=(not args.nobackup))
    finally:
        if handler:
            logger.remove_handler(handler)

    logger.blankline()

    # Keep track of errors and warnings
    errors += status[0]
    warnings += status[1]

# Report the error status if validation failed or any error occurred
if errors or (args.task == 'validate' and warnings):
    sys.exit(1)

##########################################################################################



##########################################################################################
# tests/test_asciitable.py
##########################################################################################

import pathlib
import sys
import unittest

from filecache import FCPath

from pdstemplate import PdsTemplate, TemplateError
from pdstemplate.asciitable import AsciiTable, ANALYZE_TABLE, TABLE_VALUE
from pdstemplate.asciitable import _latest_ascii_table, _reset_ascii_table


class Test_AsciiTable(unittest.TestCase):

    def runTest(self):

        self.assertIn('ANALYZE_TABLE', PdsTemplate._PREDEFINED_FUNCTIONS)
        self.assertNotIn('TABLE_VALUE', PdsTemplate._PREDEFINED_FUNCTIONS)

        root_dir = pathlib.Path(sys.modules['pdstemplate'].__file__).parent.parent
        test_file_dir = root_dir / 'test_files'
        path = test_file_dir / 'COVIMS_0094_index.tab'
        table = AsciiTable(path)
        self.assertIs(_latest_ascii_table(), table)
        self.assertEqual(table.filepath, FCPath(path))

        self.assertIn('TABLE_VALUE', PdsTemplate._PREDEFINED_FUNCTIONS)

        self.assertEqual(TABLE_VALUE('PATH'), str(path).replace('\\', '/'))
        self.assertEqual(TABLE_VALUE('BASENAME'), path.name)
        self.assertEqual(TABLE_VALUE('ROWS'), 1711)
        self.assertEqual(TABLE_VALUE('COLUMNS'), 74)
        self.assertEqual(TABLE_VALUE('ROW_BYTES'), 1089)
        self.assertEqual(TABLE_VALUE('TERMINATORS'), 2)
        self.assertEqual(TABLE_VALUE('WIDTH', 0), 27)

        with self.assertRaisesRegex(TemplateError, r'.*KeyError.*'):
            _ = TABLE_VALUE('ZZZ')

        k = 0
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', k), 'A25')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', k), '%25s')
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', k), 'CHARACTER')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', k), 'ASCII_Text_Preserved')
        self.assertEqual(TABLE_VALUE('QUOTES', k), 1)
        self.assertEqual(TABLE_VALUE('START_BYTE', k), 1)
        self.assertEqual(TABLE_VALUE('BYTES', k), 25)
        self.assertEqual(TABLE_VALUE('MINIMUM', k), "v1882439160_1.qub        ")
        self.assertEqual(TABLE_VALUE('MAXIMUM', k), "v1884114114_1.qub        ")
        self.assertEqual(TABLE_VALUE('FIRST', k), "v1882439160_1.qub        ")
        self.assertEqual(TABLE_VALUE('LAST', k), "v1884114114_1.qub        ")
        self.assertEqual(len(TABLE_VALUE('VALUES', k)), 1711)

        k = 4
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', k), 'A23')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', k), '%23s')
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', k), 'TIME')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', k), 'ASCII_Date_Time_YMD')
        self.assertEqual(TABLE_VALUE('QUOTES', k), 0)
        self.assertEqual(TABLE_VALUE('START_BYTE', k), 98)
        self.assertEqual(TABLE_VALUE('BYTES', k), 23)
        self.assertEqual(TABLE_VALUE('MINIMUM', k), "2017-08-26T10:42:54.426")
        self.assertEqual(TABLE_VALUE('MAXIMUM', k), "2017-09-14T19:58:37.608")
        self.assertEqual(TABLE_VALUE('FIRST', k), "2017-08-26T10:42:54.426")
        self.assertEqual(TABLE_VALUE('LAST', k), "2017-09-14T19:58:37.608")
        self.assertEqual(len(TABLE_VALUE('VALUES', k)), 1711)

        k = 13
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', k), 'I2')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', k), '%2d')
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', k), 'ASCII_INTEGER')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', k), 'ASCII_Integer')
        self.assertEqual(TABLE_VALUE('QUOTES', k), 0)
        self.assertEqual(TABLE_VALUE('START_BYTE', k), 275)
        self.assertEqual(TABLE_VALUE('BYTES', k), 2)
        self.assertEqual(TABLE_VALUE('MINIMUM', k), 1)
        self.assertEqual(TABLE_VALUE('MAXIMUM', k), 64)
        self.assertEqual(TABLE_VALUE('FIRST', k), 64)
        self.assertEqual(TABLE_VALUE('LAST', k), 64)
        self.assertEqual(len(TABLE_VALUE('VALUES', k)), 1711)

        k = 15
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', k), '"F10.1"')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', k), '%10.1f')
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', k), 'ASCII_REAL')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', k), 'ASCII_Real')
        self.assertEqual(TABLE_VALUE('QUOTES', k), 0)
        self.assertEqual(TABLE_VALUE('START_BYTE', k), 281)
        self.assertEqual(TABLE_VALUE('BYTES', k), 10)
        self.assertEqual(TABLE_VALUE('MINIMUM', k), 20.)
        self.assertEqual(TABLE_VALUE('MAXIMUM', k), 1000.)
        self.assertEqual(TABLE_VALUE('FIRST', k), 90.)
        self.assertEqual(TABLE_VALUE('LAST', k), 90.)
        self.assertEqual(len(TABLE_VALUE('VALUES', k)), 1711)

        k = 73
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', k), '"E11.0"')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', k), '%11.0E')
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', k), 'ASCII_REAL')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', k), 'ASCII_Real')
        self.assertEqual(TABLE_VALUE('QUOTES', k), 0)
        self.assertEqual(TABLE_VALUE('START_BYTE', k), 1077)
        self.assertEqual(TABLE_VALUE('BYTES', k), 11)
        self.assertEqual(TABLE_VALUE('MINIMUM', k), 1.e32)
        self.assertEqual(TABLE_VALUE('MAXIMUM', k), 1.e32)
        self.assertEqual(TABLE_VALUE('FIRST', k), 1.e32)
        self.assertEqual(TABLE_VALUE('LAST', k), 1.e32)
        self.assertEqual(TABLE_VALUE('VALUES', k), 1711 * [1.e32])

        # Misc errors
        self.assertRaises(ValueError, AsciiTable, path, separator='+')
        self.assertRaises(ValueError, AsciiTable, path, escape='+')

        records = [b'123\n', b'1234\n']
        self.assertRaises(TemplateError, AsciiTable, path, records)

        records = [b'12345\n', b'123,5\n']
        self.assertRaises(TemplateError, AsciiTable, path, records)

        records = [b'12345\r\n', b'23456\n']
        self.assertRaises(TemplateError, AsciiTable, path, records)

        records = [b'12345\r\n', b'234567\n']
        self.assertRaises(TemplateError, AsciiTable, path, records)

        records = [b'12345\r\n', b'234567\n']
        self.assertRaises(TemplateError, AsciiTable, path, records)

        records = [b'1234\r\n', b'2345\r\n']
        self.assertRaises(TemplateError, AsciiTable, path, records, crlf=False)

        records = [b'1234\n', b'2345\n']
        self.assertRaises(TemplateError, AsciiTable, path, records, crlf=True)

        records = [b'1234\n', b'abcd\n']
        self.assertRaises(TemplateError, AsciiTable, path, records)

        # More data types
        table = AsciiTable(path, content=[b'2017-09-14T19:58:37.608Z\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_YMD_UTC')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608Z\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_DOY_UTC')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_DOY')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608Z\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_DOY_UTC')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608 \n',
                                          b'2017-009T19:58:37.608Z\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_DOY')

        table = AsciiTable(path, content=[b'2017-09-14T19:58:37.608Z\n',
                                          b'2017-09-14T19:58:37.608 \n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_YMD')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608Z  \n',
                                          b'2017-09-14T19:58:37.608Z\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time_UTC')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608   \n',
                                          b'2017-09-14T19:58:37.608Z\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time')

        table = AsciiTable(path, content=[b'2017-009T19:58:37.608Z \n',
                                          b'2017-09-14T19:58:37.608\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_Time')

        table = AsciiTable(path, content=[b'2017-09-14\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_YMD')

        table = AsciiTable(path, content=[b'2017-009\n'])
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Date_DOY')

        table = AsciiTable(path, content=[b'123x\n'])
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'CHARACTER')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Text_Preserved')
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', 0), 'A4')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', 0), '%4s')
        self.assertEqual(TABLE_VALUE('WIDTH', 0), 4)
        self.assertEqual(TABLE_VALUE('START_BYTE', 0), 1)
        self.assertEqual(TABLE_VALUE('QUOTES', 0), 0)

        table = AsciiTable(path, content=[b'abcdef\n', b'"ABCD"\n'])
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'CHARACTER')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Text_Preserved')
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', 0), 'A6')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', 0), '%6s')
        self.assertEqual(TABLE_VALUE('WIDTH', 0), 6)
        self.assertEqual(TABLE_VALUE('START_BYTE', 0), 1)
        self.assertEqual(TABLE_VALUE('QUOTES', 0), 0)
        self.assertEqual(TABLE_VALUE('LAST', 0), '"ABCD"')

        table = AsciiTable(path, content=[b'   +1.E32\n', b'12345.678\n'])
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'ASCII_REAL')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Real')
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', 0), '"F9.3"')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', 0), '%9.3f')

        table = AsciiTable(path, content=[b'   +1.E32\n', b'12345.678\n'])
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'ASCII_REAL')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', 0), 'ASCII_Real')
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', 0), '"F9.3"')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', 0), '%9.3f')

        # Escaped quotes and quoted commas
        table = AsciiTable(path, content=[b'"123,456",789\n'])
        self.assertEqual(TABLE_VALUE('COLUMNS'), 2)
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'CHARACTER')

        self.assertRaises(TemplateError, AsciiTable, path, content=[b'"abc","def\n'])

        table = AsciiTable(path, content=[b'"123\\",456",789\n'], escape='\\')
        self.assertEqual(TABLE_VALUE('COLUMNS'), 2)
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'CHARACTER')
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', 0), 'A9')
        self.assertEqual(TABLE_VALUE('QUOTES', 0 ), 1)

        table = AsciiTable(path, content=[b'"123"",456",789\n'], escape='"')
        self.assertEqual(TABLE_VALUE('COLUMNS'), 2)
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', 0), 'CHARACTER')
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', 0), 'A9')
        self.assertEqual(TABLE_VALUE('QUOTES', 0), 1)

        table = AsciiTable(path, content=[b'"12345678\\""\n',
                                          b'"12345678  "\n',
                                          b'",,,,,,\\",,"\n'], escape='\\')
        self.assertEqual(TABLE_VALUE('FIRST', 0), '12345678"')
        self.assertEqual(TABLE_VALUE('LAST', 0), ',,,,,,",,')

        table = AsciiTable(path, content=[b'"123"\n',
                                          b'abcde\n'])
        self.assertEqual(TABLE_VALUE('FIRST', 0), '"123"')
        self.assertEqual(TABLE_VALUE('LAST', 0), 'abcde')

        # Alt separators
        table = AsciiTable(path, content=[b'"123|456"|789\n'], separator='|')
        self.assertEqual(TABLE_VALUE('FIRST', 0), '123|456')

        table = AsciiTable(path, content=[b'"123\t456"\t789\n'], separator='\t')
        self.assertEqual(TABLE_VALUE('FIRST', 0), '123\t456')

        table = AsciiTable(path, content=[b'"123;456";789\n'], separator=';')
        self.assertEqual(TABLE_VALUE('FIRST', 0), '123;456')

        # ANALYZE_TABLE
        ANALYZE_TABLE(str(path))
        self.assertEqual(TABLE_VALUE('PATH'), str(path).replace('\\', '/'))
        self.assertEqual(TABLE_VALUE('BASENAME'), path.name)
        self.assertEqual(TABLE_VALUE('ROWS'), 1711)
        self.assertEqual(TABLE_VALUE('COLUMNS'), 74)
        self.assertEqual(TABLE_VALUE('ROW_BYTES'), 1089)
        self.assertEqual(TABLE_VALUE('TERMINATORS'), 2)
        self.assertEqual(TABLE_VALUE('WIDTH', 0), 27)

        k = 0
        self.assertEqual(TABLE_VALUE('PDS3_FORMAT', k), 'A25')
        self.assertEqual(TABLE_VALUE('PDS4_FORMAT', k), '%25s')
        self.assertEqual(TABLE_VALUE('PDS3_DATA_TYPE', k), 'CHARACTER')
        self.assertEqual(TABLE_VALUE('PDS4_DATA_TYPE', k), 'ASCII_Text_Preserved')
        self.assertEqual(TABLE_VALUE('QUOTES', k), 1)
        self.assertEqual(TABLE_VALUE('START_BYTE', k), 1)
        self.assertEqual(TABLE_VALUE('BYTES', k), 25)
        self.assertEqual(TABLE_VALUE('MINIMUM', k), "v1882439160_1.qub        ")
        self.assertEqual(TABLE_VALUE('MAXIMUM', k), "v1884114114_1.qub        ")
        self.assertEqual(TABLE_VALUE('FIRST', k), "v1882439160_1.qub        ")
        self.assertEqual(TABLE_VALUE('LAST', k), "v1884114114_1.qub        ")
        self.assertEqual(len(TABLE_VALUE('VALUES', k)), 1711)

        # Reset to starting point
        del PdsTemplate._PREDEFINED_FUNCTIONS['ANALYZE_TABLE']
        del PdsTemplate._PREDEFINED_FUNCTIONS['TABLE_VALUE']
        _reset_ascii_table()



##########################################################################################
# pdstemplate/_pdsblock.py
##########################################################################################
"""Class used internally during template evaluation."""

import re
from collections import deque, namedtuple
from xml.sax.saxutils import escape

from filecache import FCPath

from .utils import TemplateError, TemplateAbort, _RaisedException
from .utils import get_logger, _NOESCAPE_FLAG

# namedtuple class definition
#
# This is used to describe any subset of lines in the template containing one header and
# any label text up to the next header:
#   header  the header type, e.g., "$FOR" or "$IF" or $END_IF";
#   arg     any expression following the header, inside parentheses;
#   line    the line number of the template in which the header appears;
#   body    the text immediately following this header and up until the next header.
#
# When the template file is first read, it is described by a deque of _Section objects. If
# there is no header before the first line of the template, it is assigned a header type
# of "$ONCE().
_Section = namedtuple('_Section', ['header', 'arg', 'line', 'body'])


class _PdsBlock(object):
    """_PdsBlock is an abstract class that describes a hierarchical section of the label
    template, beginning with a header. There are individual subclasses to support these
    different types of headers:
        _PdsForBlock     for $FOR
        _PdsIfBlock      for $IF and $ELSE_IF
        _PdsElseBlock    for $ELSE
        _PdsIncludeBlock for $INCLUDE
        _PdsNoteBlock    for $NOTE
        _PdsOnceBlock    for $END_FOR, $END_IF, $END_NOTE, and any other section of the
                         template for which what follows is included exactly once.

    Each _PdsBlock always represents a logically complete section of the template, from
    one header up to its logical completion. For example, if a template contains this
    sequence of headers:
        $FOR(...)
          $IF(...)
          $ELSE
          $END_IF
        $END_FOR
    then every line of the template from the $FOR header down to (but not including) the
    $END_FOR will be described by one _PdsForBlock. Every _PdsBlock object contains a
    "sub_block" attribute, which is a deque of all the _PdsBlocks embedded within it. In
    this case, the sub_blocks attribute will contain a single _PdsIfBlock, which in turn
    will contain a single _PdsElseBlock.

    Each _PdsBlock also has a "body" attribute, which represents the template text between
    this header and the next header. That text is pre-processed for speedier execution by
    locating all the Python expressions (surrounded by "$") embedded within it.

    The constructor for each _PdsBlock subclass takes a single deque of Sequence objects
    as input. As a side-effect, it removes one or more items from the front of the deque
    until its definition, including any nested _PdsBlocks, is complete. The constructor
    handles any nested _PdsBlocks within it by calling the constructor recursively and
    saving the results of each recursive call in its sub_blocks attribute.

    Each _PdsBlock subclass has its own execute() method. This method contains the logic
    that determines whether (for _PdsIfBlocks and _PdsElseBlocks) or how many times (for
    _PdsForBlocks) its body and sub_blocks are written into the label file. Nesting is
    handled by having each _PdsBlock call the execute method of the _PdsBlocks nested
    within it.
    """

    # This pattern matches a header record;
    #  groups(1) = line number; groups(2) = header; groups(3) = argument in parentheses
    _HEADER_WORDS = ['IF', 'ELSE_IF', 'ELSE', 'END_IF', 'FOR', 'END_FOR', 'ONCE', 'NOTE',
                     'END_NOTE', 'INCLUDE']

    # This regular expression splits up the content of the template at the location of
    # each header. For each match, it returns three groups: a leading line number, the
    # header word ("IF", "FOR", etc.), and text inside the parentheses, if any.
    _HEADER_PATTERN = re.compile(r'(?<![^\n]) *\$(\d+):(' + '|'.join(_HEADER_WORDS)
                                 + r')(\(.*\)|) *\n')

    # This pattern matches an internal assignment within an expression;
    # group(0) = variable name; group(1) = expression
    NAMED_PATTERN = re.compile(r' *([A-Za-z_]\w*) *=([^=].*)')
    ELSE_HEADERS = {'$ELSE_IF', '$ELSE', '$END_IF'}

    def preprocess_body(self):
        """Preprocess body text from the template by locating all of the embedded Python
        expressions and returning a list of substrings, where odd-numbered entries are the
        expressions to evaluate, along with the associated line number.
        """

        # Split at the "$"
        parts = self.body.split('$')
        if len(parts) % 2 != 1:
            line = parts[-1].partition(':')[0]
            raise TemplateAbort(f'Mismatched "$" at {self.filepath.name}:{line}')

        # Because we inserted the line number after every "$", every part except the first
        # now begins with a number followed by ":". We need to make the first item
        # consistent with the others.
        parts[0] = '0:' + parts[0]

        # new_parts is a deque of values that alternates between label substrings and
        # tuples (expression, name, line)

        new_parts = deque()
        for k, part in enumerate(parts):

            # Strip off the line number that we inserted after every "$"
            (line, _, part) = part.partition(':')

            # Even-numbered items are literal text
            if k % 2 == 0:
                new_parts.append(part)

            # Odd-numbered are expressions, possibly with a name
            else:

                # Look for a name
                match = _PdsBlock.NAMED_PATTERN.fullmatch(part)
                if match:
                    expression = match.group(2)
                    name = match.group(1)
                else:
                    expression = part
                    name = ''

                new_parts.append((expression, name, int(line)))

        self.preprocessed = new_parts

    def evaluate_expression(self, expression, line, state):
        """Evaluate a single expression using the state's dictionaries as needed. Identify
        the file name and line number if an error occurs.

        Parameters:
            expression (str): Expression to evaluate.
            line (int): Line number in the template starting from 1.
            state (_LabelState): State describing the label being generated.

        Returns:
            str: The evaluated expression as a string.
        """

        if expression:
            try:
                return eval(expression, state.global_dict, state.local_dicts[-1])

            # Do not pass go, do not collect $200
            except TemplateAbort:
                raise

            # This handles a call to $RAISE()
            except _RaisedException as err:
                suffix = f' at {self.filepath.name}:{line}'
                if state.raise_exceptions:
                    raise (err.exception)(err.message + suffix) from err
                get_logger().error(err.exception.__name__ + ' ' + err.message + suffix,
                                   state.label_path)
                return (f'[[[{err.exception.__name__}({err.message}){suffix}]]]')

            except Exception as err:
                # Attach the expression, file name and line number to the error message
                suffix = f' in {expression} at {self.filepath.name}:{line}'
                message = str(err) + suffix
                if state.raise_exceptions:
                    raise type(err)(message) from err

                # Log with original stacktrace
                try:
                    raise type(err)(message) from err
                except Exception as err2:
                    get_logger().exception(err2, state.label_path,
                                           more=self._more_error_info(self.line))

                # Return the content of the error message
                if isinstance(err, TemplateError):
                    return f'[[[{message}]]]'

                return f'[[[{type(err).__name__}({err})' + suffix + ']]]'

        # An empty expression is just a "$" followed by another "$"
        else:
            return '$'      # "$$" maps to "$"

    @staticmethod
    def _is_error(value):
        """True if the value is the text of an error message."""

        if not isinstance(value, str):
            return False

        return value.startswith('[[[') and value.endswith(']]]')

    def execute_body(self, state):
        """Generate the label text defined by this body, using the state's dictionaries to
        fill in the blanks. The content is returned as a deque of strings, which are to be
        joined upon completion to create the label.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate.
        """

        results = deque()
        for k, item in enumerate(self.preprocessed):

            # Even-numbered items are literal text
            if k % 2 == 0:
                results.append(item)

            # Odd-numbered items are expressions
            else:
                (expression, name, line) = item
                value = self.evaluate_expression(expression, line, state)

                if name and not _PdsBlock._is_error(value):
                    state.local_dicts[-1][name] = value

                # Format a float without unnecessary trailing zeros
                if isinstance(value, float):
                    value = _PdsBlock._pretty_truncate(value, state.template.upper_e)
                else:
                    # Otherwise, just convert to string
                    value = str(value)

                # Escape
                if self.template.xml:
                    if value.startswith(_NOESCAPE_FLAG):
                        value = value[len(_NOESCAPE_FLAG):]
                    else:
                        value = escape(value)

                results.append(value)

        return results

    def execute(self, state):
        """Evaluate this block of label text, using the dictionaries to fill in the
        blanks.

        This base class method implements the default procedure, which is to execute the
        body plus any sub-blocks exactly once. It is overridden for $FOR and $IF blocks.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate.
        """

        results = self.execute_body(state)

        for block in self.sub_blocks:
            results += block.execute(state)

        return results

    def _more_error_info(self, line):
        """The error info text to include following an exception.

        Parameters:
            line (int): Line number in template content, starting from 1.

        Returns:
            str: If template._include_more_error_info, this is the selected line of the
                template's content; otherwise, an empty string.
        """

        if not self.template._include_more_error_info:
            return ''

        recs = self.template.content.split('\n')
        return f'    {line}: ' + recs[line-1]

    ######################################################################################
    # "Compiler" from a list of template records into a deque of _PdsBlock objects
    ######################################################################################

    @staticmethod
    def process_headers(content, template, filepath=None):
        """Process the template content into a deque of _PdsBlock objects, one for each
        header found.

        Parameters:
            content (str): The entire content of the template as a single string with <LF>
                line terminators.
            template (PdsTemplate): The PdsTemplate object.
            filepath (str, Path, or FCPath, optional): Path to the source file. If not
                specified, template.template_path is used.

        Returns:
            deque[_PdsBlock]: A deque of _PdsBlock objects representing the entire content
                of the template.
        """

        # Strip inline comments
        content = _PdsBlock._strip_inline_comments(content)

        # We need to save the line number in which each expression appears so that error
        # messages can be informative. To handle this, we temporarily write the line
        # number followed by a colon after each "$" found in the template.

        # Insert line numbers after each "$"
        records = content.split('\n')
        numbered = [rec.rstrip().replace('$', f'${k+1}:')
                    for k, rec in enumerate(records)]
        content = '\n'.join(numbered)

        # Split based on headers. The entire template is split into substrings...
        # 0: text before the first header, if any
        # 1: line number of the header
        # 2: header word ("IF", "FOR", etc.)
        # 3: text between parentheses in the header line
        # 4: template text from here to the next header line
        # 5: line number of the next header
        # etc.
        parts = _PdsBlock._HEADER_PATTERN.split(content)

        # parts[0] is '' if the file begins with a header, or else it is the body text
        # before the first header. The first header is always described by parts[1:4];
        # every part indexed 4*N + 1 is a line number.

        # Create a list of (header, arg, line, body) tuples, skipping parts[0]
        sections = [_Section('$'+h, a, int(l), b) for (l, h, a, b)
                    in zip(parts[1::4], parts[2::4], parts[3::4], parts[4::4])]

        # Convert to deque and prepend the leading body text if necessary
        sections = deque(sections)
        if parts[0]:
            sections.appendleft(_Section('$ONCE', '', 0, parts[0]))

        # Convert the sections into a list of execution blocks
        # Each call to _PdsBlock.new_block pops one or more items off top of the deque;
        # the loop repeats until no sections are left.
        blocks = deque()
        while sections:
            # Each call to _PdsBlock.new_block takes as many sections off the front of the
            # deque as it needs to in order to be syntactically complete. For example, if
            # the section at the top is "IF", it will remove the subsequent "ELSE_IF" and
            # "ELSE" sections from the deque. It will return when it encounters the
            # associated "END_IF". Calls are recursive, so this handles nesting correctly.
            blocks.append(_PdsBlock.new_block(sections, template, filepath=filepath))

        return blocks

    @staticmethod
    def new_block(sections, template, filepath=None):
        """Construct an _PdsBlock subclass based on a deque of _Section tuples (header,
        arg, line,  body).

        Pop as many _Section tuples off the top of the deque as are necessary to complete
        the block and any of its internal blocks, recursively.

        Parameters:
            sections (deque[_Section]): _Section objects containing template content.
            template (PdsTemplate): The PdsTemplate object..

        Returns:
            deque[_PdsBlock]: A deque of _PdsBlock objects representing the given sections
                of the template.
        """

        filepath = FCPath(filepath) if filepath else template.template_path

        (header, arg, line, body) = sections[0]
        if header.startswith('$ONCE'):
            return _PdsOnceBlock(sections, template, filepath=filepath)
        if header.startswith('$NOTE'):
            return _PdsNoteBlock(sections, template, filepath=filepath)
        if header == '$FOR':
            return _PdsForBlock(sections, template, filepath=filepath)
        if header == '$IF':
            return _PdsIfBlock(sections, template, filepath=filepath)
        if header == '$INCLUDE':
            return _PdsIncludeBlock(sections, template, filepath=filepath)

        if header == '$END_FOR':
            raise TemplateAbort(f'$END_FOR without matching $FOR at '
                                f'{filepath.name}:{line}')
        if header == '$END_NOTE':
            raise TemplateAbort(f'$END_NOTE without matching $NOTE at '
                                f'{filepath.name}:{line}')
        if header in _PdsBlock.ELSE_HEADERS:    # pragma: no coverage - can't get here
            raise TemplateAbort(f'{header} without matching $IF at '
                                f'{filepath.name}:{line}')

        raise TemplateAbort(f'Unrecognized header {header}({arg}) at '
                            f'{filepath.name}:{line}')  # pragma: no coverage

    @staticmethod
    def _strip_inline_comments(content):
        """Strip inline comments from the given content."""

        lines = content.split('\n')
        comment = '$NOTE:'
        newlines = []
        for line in lines:
            parts = line.partition(comment)
            if parts[1]:                # if $NOTE found
                if parts[0] == '':      # remove entire line
                    continue
                newlines.append(parts[0].rstrip())
            else:
                newlines.append(line)

        return '\n'.join(newlines)

    ######################################################################################
    # Utility
    ######################################################################################

    # Modify a number if it contains ten 0's or 9's in a row, followed by other digits
    _ZEROS = re.compile(r'(.*[.1-9])0{10,99}[1-9]\d*')
    _NINES = re.compile(r'(.*\.\d+9{10,99})[0-8]\d*')

    def _pretty_truncate(value, upper_e):
        """Convert a floating-point number to a string, while suppressing any extraneous
        trailing digits by rounding to the nearest value that does not have them.

        This eliminates numbers like "1.0000000000000241" and "0.9999999999999865" in the
        label, by suppressing insignificant digits.

        Parameters:
            value (float): Value to format as a string.
            upper_e (bool): True to use uppercase "E" in exponential notation.

        Returns:
            str: Formatted string.
        """

        str_value = str(value)

        (mantissa, e, exponent) = str_value.partition('e')
        if upper_e:
            e = e.upper()

        if mantissa.endswith('.0'):
            return mantissa[:-1] + e + exponent

        if '.' not in mantissa:         # always a decimal point in the mantissa
            return mantissa + '.' + e + exponent

        # Handle trailing zeros
        match = _PdsBlock._ZEROS.fullmatch(mantissa)
        if match:
            return match.group(1) + e + exponent

        # Check for trailing nines
        match = _PdsBlock._NINES.fullmatch(mantissa)
        if not match:
            # Value looks OK; return as is
            return str_value

        # Replace every digit in the mantissa with a zero
        # This creates an string expression equal to zero, but using the exact same
        # format, including sign.
        offset_str = match.group(1)
        for c in '123456789':       # replace non-zero digits with zeros
            offset_str = offset_str.replace(c, '0')

        # Now replace the last digit with "1"
        # This is an offset (positive or negative) to zero out the trailing digits
        offset_str = offset_str[:-1] + '1'      # replace the last digit with "1"

        # Apply the offset and return
        value = float(match.group(1)) + float(offset_str)
        return str(value).rstrip('0') + e + exponent


################################################

class _PdsOnceBlock(_PdsBlock):
    """A block of text to be included once. This applies to a literal $ONCE header, and
    also to $END_FOR, $END_IF, and $END_NOTE headers."""

    WORD = r' *([A-Za-z_]\w*) *'
    PATTERN = re.compile(r'\(' + WORD + r'=([^=].*)\)')

    def __init__(self, sections, template, filepath=None):
        """Define a block to be executed once. Pop the associated sections off the stack.

        Parameters:
            sections (deque[_Section]):
                The remainder of the template's content. This constructor pops as many
                sections off the top of the deque as are needed to complete this block.
            template (PdsTemplate):
                The object being converted into _PdsBlocks.
            filepath (str, Path, or FCPath, optional):
                The file containing this $ONCE block; usually the file path of `template`
                but it could be that of an $INCLUDE file.

        Raises:
            TemplateAbort: Irrecoverable syntax error.

        Note:
            The name of a properly matched $END_IF header is changed internally to
            $ONCE-$END_IF during template initialization. Also, the name of a properly
            matched $END_FOR is changed to $ONCE-$END_FOR during template initialization,
            and $END_NOTE is changed to $ONCE-$END_NOTE. This code must strip away the
            $ONCE- prefix.
        """

        (header, arg, line, body) = sections.popleft()
        self.header = header.replace('$ONCE-', '')
        self.arg = arg
        self.name = ''
        self.line = line
        self.filepath = FCPath(filepath) if filepath else template.template_path
        self.body = body
        self.preprocess_body()
        self.sub_blocks = deque()
        self.template = template

        # Pop one entry off the local dictionary stack at the end of IF and FOR loops
        self.pop_local_dict = header in ('$ONCE-$END_FOR', '$ONCE-$END_IF')

        match = _PdsOnceBlock.PATTERN.fullmatch(arg)
        if match:
            (self.name, self.arg) = match.groups()

        if header.startswith('$ONCE-') and arg:  # pragma: no coverage
            # This can't happen in the current code because IF, FOR, and NOTE all
            # ignore the arg that's present in the template and pass in '' instead
            raise TemplateAbort(f'Extraneous argument for {self.header} at '
                                f'{self.filepath.name}:{line}')

    def execute(self, state):
        """Evaluate this block of $ONCE text unconditionally.

        Use the dictionaries to evaluate any embedded expressions.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate upon completion.
        """

        results = deque()

        # Pop the local dictionary stack if necessary
        if self.pop_local_dict:
            state.local_dicts.pop()

        # Define the local variable if necessary
        if self.arg:
            value = self.evaluate_expression(self.arg, self.line, state)
            if _PdsBlock._is_error(value):
                return deque([value])

            # Write new values into the local dictionary, not a copy
            if self.name:
                state.local_dicts[-1][self.name] = value

        # Include the body and any sub-blocks exactly once
        results += _PdsBlock.execute(self, state)
        return results


################################################

class _PdsNoteBlock(_PdsBlock):
    """A block of text between $NOTE and $END_NOTE, not to be included."""

    def __init__(self, sections, template, filepath=None):
        """Define a block to be executed zero times. Pop the associated sections off the
        stack.

        Parameters:
            sections (deque[_Section]):
                The remainder of the template's content. This constructor pops as many
                sections off the top of the deque as are needed to complete this block.
            template (PdsTemplate):
                The object being converted into _PdsBlocks.
            filepath (str, Path, or FCPath, optional):
                The file containing this $ONCE block; usually the file path of `template`
                but it could be that of an $INCLUDE file.

        Raises:
            TemplateAbort: Irrecoverable syntax error.
        """

        (header, arg, line, body) = sections.popleft()
        self.header = header
        self.arg = arg
        self.line = line
        self.filepath = FCPath(filepath) if filepath else template.template_path
        self.body = body
        self.preprocess_body()
        self.sub_blocks = deque()
        self.template = template

        if arg:
            raise TemplateAbort(f'Extraneous argument for {self.header} at '
                                f'{self.filepath.name}:{line}')

        # Save internal sub-blocks until the $END_NOTE is found
        self.sub_blocks = deque()
        while sections and sections[0].header != '$END_NOTE':
            self.sub_blocks.append(_PdsBlock.new_block(sections, template, self.filepath))

        if not sections:
            raise TemplateAbort(f'Unterminated {header} block starting at '
                                f'{self.filepath.name}:{line}')

        # Handle the matching $END_NOTE section as $ONCE
        (header, arg, line, body) = sections[0]
        sections[0] = _Section('$ONCE-' + header, '', line, body)

    def execute(self, state):
        """Ignore this block of $NOTE text.

        Use the dictionaries to evaluate any embedded expressions.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate upon completion.
        """

        return deque()


################################################

class _PdsForBlock(_PdsBlock):
    """A block of text between $FOR and $END_FOR. It is to be evaluated zero or more
    times, by iterating through the argument.
    """

    # These patterns match one, two, or three variable names, followed by "=", to be used
    # as temporary variables inside this section of the label
    WORD = r' *([A-Za-z_]\w*) *'
    PATTERN1 = re.compile(r'\(' + WORD + r'=([^=].*)\)')
    PATTERN2 = re.compile(r'\(' + WORD + ',' + WORD + r'=([^=].*)\)')
    PATTERN3 = re.compile(r'\(' + WORD + ',' + WORD + ',' + WORD + r'=([^=].*)\)')

    def __init__(self, sections, template, filepath=None):
        """Define a block to be executed inside a loop. Pop the associated section off the
        stack.

        Parameters:
            sections (deque[_Section]):
                The remainder of the template's content. This constructor pops as many
                sections off the top of the deque as are needed to complete this block.
            template (PdsTemplate):
                The object being converted into _PdsBlocks.
            filepath (str, Path, or FCPath, optional):
                The file containing this $ONCE block; usually the file path of `template`
                but it could be that of an $INCLUDE file.

        Raises:
            TemplateAbort: Irrecoverable syntax error.
        """

        (header, arg, line, body) = sections.popleft()
        self.header = header
        self.line = line
        self.filepath = FCPath(filepath) if filepath else template.template_path
        self.body = body
        self.preprocess_body()
        self.template = template

        # Interpret arg as (value=expression), (value,index=expression), etc.
        if not arg:
            raise TemplateAbort(f'Missing argument for {header} at '
                                f'{self.filepath.name}:{line}')

        self.value = 'VALUE'
        self.index = 'INDEX'
        self.length = 'LENGTH'
        self.arg = arg

        for pattern in (_PdsForBlock.PATTERN1, _PdsForBlock.PATTERN2,
                        _PdsForBlock.PATTERN3):
            match = pattern.fullmatch(arg)
            if match:
                groups = match.groups()
                self.arg = groups[-1]
                self.value = groups[0]
                if len(groups) > 2:
                    self.index = groups[1]
                if len(groups) > 3:
                    self.length = groups[2]
                break

        # Save internal sub-blocks until the $END_FOR is found
        self.sub_blocks = deque()
        while sections and sections[0].header != '$END_FOR':
            self.sub_blocks.append(_PdsBlock.new_block(sections, template, self.filepath))

        if not sections:
            raise TemplateAbort(f'Unterminated {header} block starting at '
                                f'{self.filepath.name}:{line}')

        # Handle the matching $END_FOR section as $ONCE
        (header, arg, line, body) = sections[0]
        sections[0] = _Section('$ONCE-' + header, '', line, body)

    def execute(self, state):
        """If the argument evaluates to True, execute this block of label text.

        Use the dictionaries to evaluate any embedded expressions.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate upon completion.
        """

        iterator = self.evaluate_expression(self.arg, self.line, state)
        if _PdsBlock._is_error(iterator):
            return deque([iterator])    # include the error text inside the label

        # Create a new local dictionary
        state.local_dicts.append(state.local_dicts[-1].copy())

        results = deque()
        iterator = list(iterator)
        state.local_dicts[-1][self.length] = len(iterator)
        for k, item in enumerate(iterator):
            state.local_dicts[-1][self.value] = item
            state.local_dicts[-1][self.index] = k
            results += _PdsBlock.execute(self, state)

        return results


################################################

class _PdsIfBlock(_PdsBlock):
    """A block of text beginning with $IF or $ELSE_IF and continuing to the next $ELSE_IF,
    $ELSE, or $END_IF.

    Note that subsequent $ELSE_IF sections are handled recursively.
    """

    WORD = r' *([A-Za-z_]\w*) *'
    PATTERN = re.compile(r'\(' + WORD + r'=([^=].*)\)')

    def __init__(self, sections, template, filepath=None):
        """A block to be executed conditionally.

        Parameters:
            sections (deque[_Section]):
                The remainder of the template's content. This constructor pops as many
                sections off the top of the deque as are needed to complete this block.
            template (PdsTemplate):
                The object being converted into _PdsBlocks.
            filepath (str, Path, or FCPath, optional):
                The file containing this $ONCE block; usually the file path of `template`
                but it could be that of an $INCLUDE file.

        Raises:
            TemplateAbort: Irrecoverable syntax error.
        """

        (header, arg, line, body) = sections.popleft()
        self.header = header
        self.arg = arg
        self.name = ''
        self.line = line
        self.filepath = FCPath(filepath) if filepath else template.template_path
        self.body = body
        self.preprocess_body()
        self.template = template

        if not arg:
            raise TemplateAbort(f'Missing argument for {header} at '
                                f'{self.filepath.name}:{line}')

        match = _PdsIfBlock.PATTERN.fullmatch(arg)
        if match:
            (self.name, self.arg) = match.groups()

        self.else_if_block = None
        self.else_block = None

        self.sub_blocks = deque()
        while sections and sections[0].header not in _PdsBlock.ELSE_HEADERS:
            self.sub_blocks.append(_PdsBlock.new_block(sections, template, self.filepath))

        if not sections:
            raise TemplateAbort(f'Unterminated {header} block starting at '
                                f'{self.filepath.name}:{line}')

        # Handle the first $ELSE_IF. It will handle more $ELSE_IFs and $ELSEs recursively.
        if sections[0].header == '$ELSE_IF':
            self.else_if_block = _PdsIfBlock(sections, template)
            return

        # Handle $ELSE
        if sections[0].header == '$ELSE':
            self.else_block = _PdsElseBlock(sections, template)
            return

        # Handle the matching $END_IF section as $ONCE
        (header, arg, line, body) = sections[0]
        sections[0] = _Section('$ONCE-' + header, '', line, body)

    def execute(self, state):
        """If the argument evaluates to True, execute this block of label text. Otherwise,
        continue with the next $ELSE_IF or $ELSE block.

        Use the dictionaries to evaluate any embedded expressions.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate upon completion.
        """

        status = self.evaluate_expression(self.arg, self.line, state)
        if _PdsBlock._is_error(status):
            return deque([status])      # include the error text inside the label

        # Create a new local dictionary for IF but not ELSE_IF
        if self.header == '$IF':
            state.local_dicts.append(state.local_dicts[-1].copy())

        if self.name:
            state.local_dicts[-1][self.name] = status

        if status:
            return _PdsBlock.execute(self, state)

        elif self.else_if_block:
            return self.else_if_block.execute(state)

        elif self.else_block:
            return self.else_block.execute(state)

        else:
            return deque()  # empty response


################################################

class _PdsElseBlock(_PdsBlock):

    def __init__(self, sections, template, filepath=None):
        """A block to be executed only if all preceding $IF and $ELSE_IF blocks have not
        executed.

        Parameters:
            sections (deque[_Section]):
                The remainder of the template's content. This constructor pops as many
                sections off the top of the deque as are needed to complete this block.
            template (PdsTemplate):
                The object being converted into _PdsBlocks.
            filepath (str, Path, or FCPath, optional):
                The file containing this $ONCE block; usually the file path of `template`
                but it could be that of an $INCLUDE file.

        Raises:
            TemplateAbort: Irrecoverable syntax error.
        """

        (header, arg, line, body) = sections.popleft()
        self.header = header
        self.arg = arg
        self.line = line
        self.filepath = FCPath(filepath) if filepath else template.template_path
        self.body = body
        self.preprocess_body()
        self.template = template

        # Save internal sub-blocks until the $END_IF is found
        self.sub_blocks = deque()
        while sections and sections[0].header != '$END_IF':
            self.sub_blocks.append(_PdsBlock.new_block(sections, template, self.filepath))

        if not sections:
            raise TemplateAbort(f'Unterminated {header} block starting at '
                                f'{self.filepath.name}:{line}')

        # Handle the matching $END_IF section as $ONCE
        (header, arg, line, body) = sections[0]
        sections[0] = _Section('$ONCE-' + header, '', line, body)

################################################

class _PdsIncludeBlock(_PdsBlock):
    """A reference to an external file to be included at this location of the template.

    It is followed by a standard block of text to be executed once.
    """

    def __init__(self, sections, template, filepath=None):
        """A directive to include text from an specified file and then compile and
        execute it.

        Parameters:
            sections (deque[_Section]):
                The remainder of the template's content. This constructor pops as many
                sections off the top of the deque as are needed to complete this block.
            template (PdsTemplate):
                The object being converted into _PdsBlocks.
            filepath (str, Path, or FCPath, optional):
                The file containing this $ONCE block; usually the file path of `template`
                but it could be that of an $INCLUDE file.

        Raises:
            TemplateAbort: Irrecoverable syntax error.
        """

        (header, arg, line, body) = sections.popleft()
        self.header = header
        self.arg = arg
        self.name = ''
        self.line = line
        self.filepath = FCPath(filepath) if filepath else template.template_path
        self.body = body
        self.preprocess_body()
        self.sub_blocks = deque()
        self.template = template

        if not arg:
            raise TemplateAbort(f'Missing argument for {header} at '
                                f'{self.filepath.name}:{line}')

    def execute(self, state):
        """Read, compile, and execute the specified file, followed by the remaining body
        text.

        Use the dictionaries to evaluate any embedded expressions.

        Parameters:
            state (_LabelState): State describing the label being generated.

        Returns:
            deque[str]: Deque of strings to concatenate upon completion.
        """

        results = deque()

        # Interpret the file name
        filename = self.evaluate_expression(self.arg, self.line, state)
        if _PdsBlock._is_error(filename):
            return deque(['$INCLUDE(', filename, ')\n'])  # put error text into the label

        # Read the file
        try:
            content = _PdsIncludeBlock.get_content(filename,
                                                   self.template._include_dirs())
        except Exception as err:
            message = f'{repr(err)} in $INCLUDE at {state.filepath.name}:{self.line}'
            if state.raise_exceptions:
                raise type(err)(message) from err
            try:
                raise type(err)(message) from err
            except Exception as err:
                get_logger().exception(err, state.label_path,
                                       more=self._more_error_info(self.line))
            return deque(['$INCLUDE(', filename, ')\n'])  # put error text into the label

        # Compile and execute the included template
        blocks = _PdsBlock.process_headers(content, self.template, filepath=self.filepath)
        for block in blocks:
            results += block.execute(state)

        # Include the body and any sub-blocks afterward
        results += _PdsBlock.execute(self, state)
        return results

    @staticmethod
    def get_content(filename, include_dirs):
        """The content of the specified include file.

        Parameters:
            filename (str, Path, or FCPath): The name or path to the file to include.
            include_dirs (list[Path or FCPath): Ordered list of directories in which to
                look for the named file.

        Returns:
            str: The content of the file as a single string containing "\n" line
                terminators.

        Raises:
            FileNotFoundError: If the file is not found.
            OSError: Any subclass of OSError explaining why the file could not be read.
        """

        # First try to read the file directly
        filepath = FCPath(filename)
        try:
            return filepath.read_text()     # convert <CR><LF> to <LF>
        except FileNotFoundError:
            pass

        # Try each directory in turn
        for dir in include_dirs:
            try:
                return (dir / filename).read_text()
            except (FileNotFoundError, NotImplementedError):
                pass

        # On failure, re-raise the first exception
        return filepath.read_text()

##########################################################################################



##########################################################################################
# tests/test_pds3table.py
##########################################################################################

from contextlib import redirect_stdout
import io
import os
import pathlib
import re
import shutil
import sys
import tempfile
import time
import unittest

import pdslogger

from pdstemplate import PdsTemplate, get_logger, TemplateError
from pdstemplate.pds3table import Pds3Table, VALIDATE_PDS3_LABEL, \
                                  LABEL_VALUE, OLD_LABEL_VALUE, ANALYZE_TABLE
from pdstemplate.pds3table import pds3_table_preprocessor, _latest_pds3_table
from pdstemplate.asciitable import AsciiTable, _reset_ascii_table


class Test_Pds3Table(unittest.TestCase):

    def runTest(self):
        # Reset to starting point in case test_asciitable failed
        try:
            del PdsTemplate._PREDEFINED_FUNCTIONS['ANALYZE_TABLE']
        except KeyError:
            pass
        try:
            del PdsTemplate._PREDEFINED_FUNCTIONS['TABLE_VALUE']
        except KeyError:
            pass
        _reset_ascii_table()

        # No logging to stdout
        _LOGGER = get_logger()
        _LOGGER.add_handler(pdslogger.NULL_HANDLER)

        root_dir = pathlib.Path(sys.modules['pdstemplate'].__file__).parent.parent
        test_file_dir = root_dir / 'test_files'
        path = test_file_dir / 'COVIMS_0094_index.lbl'
        label = Pds3Table(path)

        self.assertRaisesRegex(TemplateError, r'No ASCII table has been.*',
                               label.assign_to)

        self.assertEqual(OLD_LABEL_VALUE('RECORD_TYPE'), 'FIXED_LENGTH')
        self.assertEqual(OLD_LABEL_VALUE('RECORD_BYTES'), 1089)
        self.assertEqual(OLD_LABEL_VALUE('FILE_RECORDS'), 1711)
        self.assertEqual(OLD_LABEL_VALUE('INTERCHANGE_FORMAT'), 'ASCII')
        self.assertEqual(OLD_LABEL_VALUE('ROWS'), 1711)
        self.assertEqual(OLD_LABEL_VALUE('COLUMNS'), 61)
        self.assertEqual(OLD_LABEL_VALUE('ROW_BYTES'), 1089)

        self.assertEqual(LABEL_VALUE('PATH'), str(path).replace('\\', '/'))
        self.assertEqual(LABEL_VALUE('BASENAME'), path.name)
        self.assertEqual(LABEL_VALUE('RECORD_TYPE'), 'FIXED_LENGTH')
        self.assertEqual(LABEL_VALUE('INTERCHANGE_FORMAT'), 'ASCII')
        self.assertEqual(LABEL_VALUE('TABLE_PATH'),
                         str(test_file_dir / 'COVIMS_0094_index.tab').replace('\\', '/'))
        self.assertEqual(LABEL_VALUE('TABLE_BASENAME'), 'COVIMS_0094_index.tab')

        ANALYZE_TABLE(_latest_pds3_table().get_table_path())

        self.assertEqual(LABEL_VALUE('RECORD_BYTES'), 1089)
        self.assertEqual(LABEL_VALUE('FILE_RECORDS'), 1711)
        self.assertEqual(LABEL_VALUE('ROWS'), 1711)
        self.assertEqual(LABEL_VALUE('COLUMNS'), 61)
        self.assertEqual(LABEL_VALUE('ROW_BYTES'), 1089)

        k = 1
        self.assertEqual(OLD_LABEL_VALUE('NAME', k), 'FILE_NAME')
        self.assertEqual(OLD_LABEL_VALUE('DATA_TYPE', k), 'CHARACTER')
        self.assertEqual(OLD_LABEL_VALUE('START_BYTE', k), 2)
        self.assertEqual(OLD_LABEL_VALUE('BYTES', k), 25)
        self.assertEqual(OLD_LABEL_VALUE('FORMAT', k), None)
        self.assertEqual(OLD_LABEL_VALUE('NOT_APPLICABLE_CONSTANT', k), None)
        self.assertEqual(OLD_LABEL_VALUE('COLUMN_NUMBER', k), None)

        self.assertEqual(LABEL_VALUE('NAME', k), 'FILE_NAME')
        self.assertEqual(LABEL_VALUE('DATA_TYPE', k), 'CHARACTER')
        self.assertEqual(LABEL_VALUE('START_BYTE', k), 2)
        self.assertEqual(LABEL_VALUE('BYTES', k), 25)
        self.assertEqual(LABEL_VALUE('FORMAT', k), 'A25')
        self.assertEqual(LABEL_VALUE('NOT_APPLICABLE_CONSTANT', k), None)
        self.assertEqual(LABEL_VALUE('COLUMN_NUMBER', k), 1)

        k = 'CENTRAL_BODY_DISTANCE'
        self.assertEqual(OLD_LABEL_VALUE('NAME', k), 'CENTRAL_BODY_DISTANCE')
        self.assertEqual(OLD_LABEL_VALUE('DATA_TYPE', k), 'ASCII_REAL')
        self.assertEqual(OLD_LABEL_VALUE('START_BYTE', k), 388)
        self.assertEqual(OLD_LABEL_VALUE('BYTES', k), 14)
        self.assertEqual(OLD_LABEL_VALUE('FORMAT', k), 'E14')
        self.assertEqual(OLD_LABEL_VALUE('NOT_APPLICABLE_CONSTANT', k), -1.e32)
        self.assertEqual(OLD_LABEL_VALUE('NULL_CONSTANT', k), 1.e32)
        self.assertEqual(OLD_LABEL_VALUE('COLUMN_NUMBER', k), None)

        self.assertEqual(LABEL_VALUE('NAME', k), 'CENTRAL_BODY_DISTANCE')
        self.assertEqual(LABEL_VALUE('DATA_TYPE', k), 'ASCII_REAL')
        self.assertEqual(LABEL_VALUE('START_BYTE', k), 388)
        self.assertEqual(LABEL_VALUE('BYTES', k), 14)
        self.assertEqual(LABEL_VALUE('FORMAT', k), 'E14')
        self.assertEqual(LABEL_VALUE('NOT_APPLICABLE_CONSTANT', k), -1.e32)
        self.assertEqual(LABEL_VALUE('NULL_CONSTANT', k), 1.e32)
        self.assertEqual(LABEL_VALUE('COLUMN_NUMBER', k), 24)
        self.assertEqual(LABEL_VALUE('MINIMUM_VALUE', k), 1.e32)
        self.assertEqual(LABEL_VALUE('MAXIMUM_VALUE', k), 1e+32)

        # Validation, edits
        warnings = label._validation_warnings()
        answer = ['SWATH_WIDTH:DATA_TYPE error: INTEGER -> ASCII_INTEGER',
                  'SWATH_LENGTH:DATA_TYPE error: INTEGER -> ASCII_INTEGER',
                  'IR_EXPOSURE:DATA_TYPE error: REAL -> ASCII_REAL',
                  'VIS_EXPOSURE:DATA_TYPE error: REAL -> ASCII_REAL']
        self.assertEqual(warnings, answer)

        label = Pds3Table(path, edits=['SC_SUN_POSITION_VECTOR:UNKNOWN_CONSTANT = 0.',
                                       'SC_SUN_VELOCITY_VECTOR:NULL_CONSTANT = 0.'])
        warnings = label._validation_warnings()
        self.assertEqual(warnings[:4], answer[:4])
        self.assertEqual(warnings[-2:],
                    ['SC_SUN_POSITION_VECTOR:UNKNOWN_CONSTANT was inserted: 0.',
                     'SC_SUN_VELOCITY_VECTOR:NULL_CONSTANT was edited: 1.E+32 -> 0.'])

        self.assertEqual(LABEL_VALUE('MINIMUM_VALUE', 'SC_SUN_POSITION_VECTOR'), 0.)
        self.assertEqual(LABEL_VALUE('MAXIMUM_VALUE', 'SC_SUN_POSITION_VECTOR'), 0.)
        before = [_LOGGER.message_count('error'), _LOGGER.message_count('warn')]
        VALIDATE_PDS3_LABEL()
        after = [_LOGGER.message_count('error'), _LOGGER.message_count('warn')]
        self.assertEqual(after[0] - before[0], 0)
        self.assertEqual(after[1] - before[1], 6)

        # More options
        label = Pds3Table(path, numbers=True, formats=True, minmax='float',
                          derived=('SC_SUN_POSITION_VECTOR', 'int'),
                          edits=['SC_SUN_POSITION_VECTOR:UNKNOWN_CONSTANT = 0.',
                                 'SC_SUN_VELOCITY_VECTOR:NULL_CONSTANT = 0.'])
        with (test_file_dir / 'COVIMS_0094_index_template.txt').open('rb') as f:
            answer = f.read().decode('latin-1')
        # print(label.content)
        self.assertEqual(answer, label.content)

        # Preprocessor
        kwargs = {'validate': True, 'numbers': True, 'formats': True, 'minmax': 'float',
                  'derived': ('SC_SUN_POSITION_VECTOR', 'int'),
                  'edits': ['SC_SUN_POSITION_VECTOR:UNKNOWN_CONSTANT = 0.',
                            'SC_SUN_VELOCITY_VECTOR:NULL_CONSTANT = 0.']}
        template = PdsTemplate(path, preprocess=pds3_table_preprocessor, kwargs=kwargs,
                               crlf=True, upper_e=True)

        dirpath = pathlib.Path(tempfile.mkdtemp())
        try:
            outpath = dirpath / 'test.lbl'
            tablepath = dirpath / 'test.tab'
            tablepath.symlink_to(test_file_dir / 'COVIMS_0094_index.tab')

            # Save...
            template.write({}, outpath, mode='save')
            with outpath.open('rb') as f:
                written = f.read().decode('latin-1')
            with (test_file_dir / 'COVIMS_0094_index_test1.txt').open('rb') as f:
                answer = f.read().decode('latin-1')
            # print(written)
            self.assertEqual(written, answer)

            # Repair does nothing
            mtime = os.path.getmtime(outpath)
            time.sleep(1)
            template.write({}, outpath, mode='repair')
            self.assertEqual(mtime, os.path.getmtime(outpath))  # file unchanged

            # Save creates backup
            template.write({}, outpath, mode='save', backup=True)
            all_paths = list(outpath.parent.iterdir())
            paths = list(all_paths)
            paths.remove(outpath)
            paths.remove(tablepath)
            self.assertIsNotNone(re.fullmatch(r'.*/test_20\d\d-\d\d-\d\dT.*\.lbl',
                                              paths[0].as_posix()))
            mtime = os.path.getmtime(outpath)

            # Validate
            F = io.StringIO()           # capture stdout to a string
            with redirect_stdout(F):
                template.write({}, outpath, mode='validate')
            _ = F.getvalue()

            self.assertEqual(mtime, os.path.getmtime(outpath))  # file unchanged

            paths = list(all_paths)
            paths.remove(outpath)

            self.assertRaises(ValueError, template.write, {}, outpath, mode='xxx')

        finally:
            shutil.rmtree(dirpath)

        # Test of sky_summary template
        kwargs = {'validate': False, 'numbers': True, 'formats': True}
        sky_summary = PdsTemplate(test_file_dir / 'sky_summary_template.txt',
                                  preprocess=pds3_table_preprocessor, kwargs=kwargs,
                                  crlf=True, upper_e=True)

        self.maxDiff = None
        with open(test_file_dir/'sky_summary_template_preprocessed.txt', 'r') as f:
            answer = f.read()
        self.assertEqual(sky_summary.content, answer)

        label = sky_summary.generate({}, test_file_dir / 'GO_0023_sky_summary.lbl')
        with open(test_file_dir/'GO_0023_sky_summary.lbl', 'rb') as f:
            answer = f.read()
        answer = answer.decode('utf-8')

        self.assertEqual(label, answer)

        # Missing TABLE
        template_path = test_file_dir / 'sky_summary_template.txt'
        with template_path.open('rb') as f:
            content = f.read()
        content = content.decode('utf-8')
        content = content.replace('TABLE', 'IMAGE')
        self.assertRaisesRegex(TemplateError, r'Template does not contain.*',
                               PdsTemplate, template_path, content,
                               preprocess=pds3_table_preprocessor,
                               crlf=True, upper_e=True)

        # Two tables
        with template_path.open('rb') as f:
            content = f.read()
        content = content.decode('utf-8')
        content = content + content
        self.assertRaisesRegex(TemplateError, r'Template contains multiple.*',
                               PdsTemplate, template_path, content,
                               preprocess=pds3_table_preprocessor,
                               crlf=True, upper_e=True)

        # Repeated column names
        path = test_file_dir / 'repeated_column_names.lbl'
        label = Pds3Table(path)
        _ = AsciiTable(test_file_dir / 'GO_0023_sky_summary.tab')
        warnings = label._validation_warnings()
        self.assertEqual(warnings,
                ['ERROR: Name FILE_SPECIFICATION_NAME is duplicated at columns 2 and 3',
                 'ERROR: Name MAXIMUM_DECLINATION is duplicated at columns 4 and 5',
                 'ERROR: Name MAXIMUM_DECLINATION is duplicated at columns 5 and 6'])

        # Reset to starting point
        del PdsTemplate._PREDEFINED_FUNCTIONS['ANALYZE_PDS3_LABEL']
        del PdsTemplate._PREDEFINED_FUNCTIONS['VALIDATE_PDS3_LABEL']
        del PdsTemplate._PREDEFINED_FUNCTIONS['LABEL_VALUE']
        del PdsTemplate._PREDEFINED_FUNCTIONS['ANALYZE_TABLE']
        del PdsTemplate._PREDEFINED_FUNCTIONS['TABLE_VALUE']

        _LOGGER.remove_all_handlers()
